"""CIF syntax highlighting for the text editor.

Copyright (c) 2025, Daniel N. Rainer (ORCID: 0000-0002-3272-3161)
All rights reserved.

BSD 3-Clause License

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


* Almot comlete redesign by Daniel Kratzert from original Idea by Daniel N. Rainer
"""

import re

from qtpy.QtGui import QTextCharFormat, QSyntaxHighlighter, QColor, QFont


class CIFSyntaxHighlighter(QSyntaxHighlighter):

    MULTILINE = 1
    LOOP_FIELDS = 2
    LOOP_DATA = 3

    def __init__(self, parent=None):
        super().__init__(parent)

        # ---------- Formats ----------

        self.bold_format = QTextCharFormat()
        self.bold_format.setFontWeight(QFont.Weight.Bold)

        self.field_format = QTextCharFormat()
        self.field_format.setForeground(QColor("#0000FF"))

        self.value_format = QTextCharFormat()
        self.value_format.setForeground(QColor("#008000"))

        self.multiline_format = QTextCharFormat()
        self.multiline_format.setForeground(QColor("#800080"))

        self.loop_keyword_format = QTextCharFormat()
        self.loop_keyword_format.setForeground(QColor("#FF6600"))
        self.loop_keyword_format.setFontWeight(QFont.Weight.Bold)

        self.loop_field_format = QTextCharFormat()
        self.loop_field_format.setForeground(QColor("#CC6600"))

        self.loop_values_format = QTextCharFormat()
        self.loop_values_format.setForeground(QColor("#996600"))

        self.vrf_values_format = QTextCharFormat()
        self.vrf_values_format.setForeground(QColor("#8b0000"))
        self.vrf_values_format.setFontWeight(QFont.Weight.Bold)

        self.field_re = re.compile(r'^_[A-Za-z][A-Za-z0-9_.\-\[\]()/]*')
        self.quoted_re = re.compile(r"'[^']*'")

    def highlightBlock(self, text: str) -> None:
        prev_state = self.previousBlockState()
        in_multiline = prev_state == self.MULTILINE
        in_loop_fields = prev_state == self.LOOP_FIELDS
        in_loop_data = prev_state == self.LOOP_DATA

        stripped = text.strip()
        lower = stripped.lower()

        # ---------- Multiline text blocks ----------

        if text.startswith(';'):
            self.setFormat(0, 1, self.bold_format)
            # self.setFormat(0, len(text), self.multiline_format)

            if in_multiline:
                self.setCurrentBlockState(0)
            else:
                self.setCurrentBlockState(self.MULTILINE)
            return

        if in_multiline:
            # This line is quite slow:
            # self.setFormat(0, len(text), self.multiline_format)
            self.setCurrentBlockState(self.MULTILINE)
            return

        # ---------- Loop start ----------

        if lower == "loop_":
            self.setFormat(0, len(text), self.loop_keyword_format)
            self.setCurrentBlockState(self.LOOP_FIELDS)
            return

        # ---------- Loop handling ----------

        if in_loop_fields or in_loop_data:
            if lower.startswith(("data_", "save_", "global_", "stop_")):
                in_loop_fields = False
                in_loop_data = False
                self.setCurrentBlockState(0)
            elif stripped.startswith('_') and not in_loop_data:
                self.setFormat(0, len(text), self.loop_field_format)
                self.setCurrentBlockState(self.LOOP_FIELDS)
                return
            elif stripped and not stripped.startswith('_'):
                in_loop_data = True
                self.setFormat(0, len(text), self.loop_values_format)
                self.setCurrentBlockState(self.LOOP_DATA)
                return
            elif not stripped and in_loop_data:
                self.setCurrentBlockState(0)
                return

        # ---------- Data tags ----------

        if text.startswith("data_"):
            self.setFormat(0, len(text), self.bold_format)

        # ---------- Field names ----------

        if text.startswith('_'):
            m = self.field_re.match(text)
            if m:
                self.setFormat(m.start(), m.end() - m.start(), self.field_format)

        if text.startswith('_vrf'):
            self.setFormat(0, len(text), self.vrf_values_format)

        # ---------- Quoted values ----------

        if "'" in text:
            for m in self.quoted_re.finditer(text):
                self.setFormat(m.start(), m.end() - m.start(), self.value_format)

        self.setCurrentBlockState(0)



import os
import unittest
from tests.helpers import AppWindowTestCase
from pathlib import Path
from unittest.mock import Mock

import docx
from docx import Document
from docx.enum.shape import WD_INLINE_SHAPE
from docx.shape import InlineShapes
from docx.shared import Cm
from docx.table import Table
from packaging.version import Version

from finalcif.appwindow import AppWindow
from finalcif.cif.cif_file_io import CifContainer
from finalcif.report.tables import make_report_from

os.environ["RUNNING_TEST"] = 'True'
data = Path('tests')
test_data = Path('test-data')


class TablesTestCase(AppWindowTestCase):

    def setUp(self) -> None:
        self.testcif = (data / 'examples/1979688_small.cif').absolute()
        self.app = AppWindow(file=self.testcif)
        self.app.ui.HAtomsCheckBox.setChecked(False)
        self.app.ui.ReportTextCheckBox.setChecked(False)
        self.app.ui.PictureWidthDoubleSpinBox.setValue(0.0)
        # make sure to use no template:
        self.app.ui.docxTemplatesListWidget.setCurrentRow(0)
        # self.app.show()
        self.reportdoc = self.app.cif.finalcif_file_prefixed(prefix='report_', suffix='-finalcif.docx')
        self.report_zip = self.app.cif.finalcif_file_prefixed(prefix='', suffix='-finalcif.zip')
        self.app.ui.PictureWidthDoubleSpinBox.setValue(7.43)
        self.app.select_report_picture(Path('finalcif/icon/finalcif.png'))

    def tearDown(self) -> None:
        self.app.cif.finalcif_file.unlink(missing_ok=True)
        self.reportdoc.unlink(missing_ok=True)
        self.report_zip.unlink(missing_ok=True)
        self.app.ui.ReportTextCheckBox.setChecked(False)
        self.app.ui.HAtomsCheckBox.setChecked(False)
        self.app.ui.PictureWidthDoubleSpinBox.setValue(7.5)
        self.app.close()
        super().tearDown()

    def test_picture_has_correct_size(self):
        self.app.ui.SaveFullReportButton.click()
        doc = Document(self.reportdoc.absolute())
        shapes: InlineShapes = doc.inline_shapes
        self.assertEqual(WD_INLINE_SHAPE.PICTURE, shapes[0].type)
        self.assertEqual(Cm(7.43).emu, shapes[0].width)

    def test_default_picture_width(self):
        self.app.ui.PictureWidthDoubleSpinBox.setValue(0.0)
        self.app.ui.SaveFullReportButton.click()
        doc = Document(self.reportdoc.resolve())
        shapes: InlineShapes = doc.inline_shapes
        self.assertEqual(WD_INLINE_SHAPE.PICTURE, shapes[0].type)
        self.assertEqual(Cm(7.5).emu, shapes[0].width)


class TemplateReportWithoutAppTestCase(AppWindowTestCase):
    def setUp(self) -> None:
        self.testcif = (data / 'examples/1979688.cif').absolute()
        self.cif = CifContainer(self.testcif)
        self.options = Mock()
        self.options.picture_width = 7.43
        self.options.without_h = False
        self.text_template = Path('finalcif/template/template_text.docx').absolute()
        self.template_without_text = Path('finalcif/template/template_without_text.docx').absolute()
        self.reportdoc = self.cif.finalcif_file_prefixed(prefix='report_', suffix='-finalcif.docx')
        self.report_zip = self.cif.finalcif_file_prefixed(prefix='', suffix='-finalcif.zip')
        self.report_pic = Path('finalcif/icon/finalcif.png')

    def tearDown(self) -> None:
        self.reportdoc.unlink(missing_ok=True)
        self.report_zip.unlink(missing_ok=True)

    def test_option_with_h(self):
        make_report_from(options=self.options, cif=self.cif,
                         output_filename=str(self.reportdoc), picfile=self.report_pic)
        doc = Document(self.reportdoc.absolute())
        table: Table = doc.tables[3]
        self.assertEqual('C1–H1', table.cell(row_idx=4, col_idx=0).text)

    def test_option_without_h(self):
        self.options.without_h = True
        make_report_from(options=self.options, cif=self.cif,
                         output_filename=str(self.reportdoc), picfile=self.report_pic)
        doc = Document(self.reportdoc.absolute())
        table: Table = doc.tables[3]
        self.assertEqual('O1–C13', table.cell(row_idx=4, col_idx=0).text)

    def test_all_paragraphs(self):
        make_report_from(options=self.options, cif=self.cif,
                         output_filename=str(self.reportdoc), picfile=self.report_pic)
        doc = Document(self.reportdoc.absolute())
        newline = '\n' if Version(docx.__version__) < Version('1.0') else ''
        result = ('Structure Tables\n'
                  '\n'
                  '\n'
                  '\n'
                  f'{newline}Table 1. Crystal data and structure refinement for cu_BruecknerJK_153F40_0m\n{newline}'
                  '\n'
                  '\n'
                  'Refinement details for cu_BruecknerJK_153F40_0m\n'
                  'The methanol molecule is disordered around a special position and thus half '
                  'occupied.\n'
                  'Table 2. Atomic coordinates and Ueq\xa0[Å2] for cu_BruecknerJK_153F40_0m\n'
                  'Ueq is defined as 1/3 of the trace of the orthogonalized Uij tensor.\n'
                  '\n'
                  'Table 3. Anisotropic displacement parameters [Å2] for '
                  'cu_BruecknerJK_153F40_0m.\n'
                  'The anisotropic displacement factor exponent takes the form: −2π2[\u205f'
                  'h2(a*)2U11\u205f+\u205fk2(b*)2U22\u205f+\u205f…\u205f+\u205f2hka*b*U12\u205f'
                  ']\n'
                  'Table 4. Bond lengths and angles for cu_BruecknerJK_153F40_0m\n'
                  '\n'
                  '\n'
                  'Table 5. Torsion angles for cu_BruecknerJK_153F40_0m\n'
                  '\n'
                  '\n'
                  '\n'
                  'Bibliography'
                  )

        self.assertEqual(result, '\n'.join([x.text for x in doc.paragraphs]))


class TablesNoPictureTestCase(AppWindowTestCase):
    def setUp(self) -> None:
        self.testcif = (data / 'examples/1979688.cif').absolute()
        self.cif = CifContainer(self.testcif)
        self.options = Mock()
        self.options.picture_width = 7.43
        self.options.without_h = False
        self.text_template = Path('finalcif/template/template_text.docx').absolute()
        self.template_without_text = Path('finalcif/template/template_without_text.docx').absolute()
        self.reportdoc = self.cif.finalcif_file_prefixed(prefix='report_', suffix='-finalcif.docx')
        self.report_zip = self.cif.finalcif_file_prefixed(prefix='', suffix='-finalcif.zip')
        self.report_pic = Path('finalcif/icon/finalcif.png')

    def tearDown(self) -> None:
        self.reportdoc.unlink(missing_ok=True)
        self.report_zip.unlink(missing_ok=True)

    def test_save_report_works(self):
        make_report_from(options=self.options, cif=self.cif,
                         output_filename=str(self.reportdoc), picfile=None)
        self.assertEqual(True, self.reportdoc.exists())

    def test_picture_has_correct_size(self):
        make_report_from(options=self.options, cif=self.cif,
                         output_filename=str(self.reportdoc), picfile=None)
        doc = Document(self.reportdoc.resolve())
        shapes: InlineShapes = doc.inline_shapes
        self.assertEqual(0, len(shapes))

    def test_picture_shape_exists(self):
        make_report_from(options=self.options, cif=self.cif,
                         output_filename=str(self.reportdoc), picfile=self.report_pic)
        doc = Document(self.reportdoc.resolve())
        shapes: InlineShapes = doc.inline_shapes
        self.assertEqual(1, len(shapes))


class ReportWithsymmetryTestCase(AppWindowTestCase):
    def setUp(self) -> None:
        self.testcif = (test_data / 'p31c.cif').resolve()
        self.cif = CifContainer(self.testcif)
        self.options = Mock()
        self.options.without_h = False
        self.reportdoc = self.cif.finalcif_file_prefixed(prefix='report_', suffix='-finalcif.docx')

    def tearDown(self) -> None:
        self.reportdoc.unlink(missing_ok=True)
        self.cif.finalcif_file.unlink(missing_ok=True)

    def test_symmetry_indicators(self):
        make_report_from(options=self.options, cif=self.cif, output_filename=str(self.reportdoc), picfile=None)
        doc = Document(self.reportdoc.absolute())
        table: Table = doc.tables[3]
        # Bond:
        self.assertEqual('C2–C3#1', table.cell(row_idx=18, col_idx=0).text)
        # Angle:
        self.assertEqual("C3'#1–C2'–C3'", table.cell(row_idx=148, col_idx=0).text)
        # Torsion angle:
        table: Table = doc.tables[4]
        self.assertEqual("C3#1–C2–C3–N1", table.cell(row_idx=6, col_idx=0).text)
        # Hydrogen bond:
        table: Table = doc.tables[5]
        self.assertEqual("N1^a–H1^a⋯Cl1#1", table.cell(row_idx=1, col_idx=0).text)


if __name__ == '__main__':
    unittest.main()



import ctypes
import os
import sys
from pathlib import Path

from finalcif import VERSION
from qtpy import QtCore, compat
from qtpy.QtWidgets import QMessageBox, QMainWindow, QVBoxLayout, QTextEdit, QPushButton, QFrame


def do_update_program(version) -> None:
    updater_exe = str(Path(__file__).parent.parent.parent.joinpath('update.exe'))
    args = ['-v', version,
            '-p', 'finalcif']
    # Using this, because otherwise I can not write to the program dir:
    ctypes.windll.shell32.ShellExecuteW(None, "runas", updater_exe, " ".join(args), None, 1)


def unable_to_open_message(parent, filepath: Path, not_ok: Exception) -> None:
    """
    Shows a message if the current cif file can not be opened.
    """
    if "PYTEST_CURRENT_TEST" in os.environ:
        print('DBG> Running inside a pytest -> not showing error message.')
        return
    info = QMessageBox(parent=parent)
    info.setIcon(QMessageBox.Icon.Information)
    print('Output from gemmi:', not_ok)
    try:
        line = str(not_ok)[4:].split(':')[1]
    except IndexError:
        line = None
    info.setText('This cif file is not readable!                                           ')
    if line:
        try:
            int(line)
            info.setInformativeText(f'\nPlease check line {line} in\n{filepath.name}')
        except ValueError:
            info.setInformativeText(f'"{filepath.name}"\n{not_ok}')
    else:
        info.setInformativeText(f'"{filepath.name}"\n{not_ok}')
    info.setModal(True)
    info.show()


def show_res_checksum_warning(parent) -> None:
    """
    A message box to display if the checksums do not agree.
    """
    if "PYTEST_CURRENT_TEST" in os.environ:
        print('DBG> Running inside a pytest -> not showing error message.')
        return
    info = QMessageBox(parent=parent)
    info.setIcon(QMessageBox.Icon.Warning)
    info.setText('The "_shelx_res_checksum" is not consistent with the .res file content!\n\n'
                 'This error might originate from non-ascii Characters like Umlauts in you SHELX file.')
    info.setModal(True)
    info.show()


def show_hkl_checksum_warning(parent) -> None:
    """
    A message box to display if the checksums do not agree.
    """
    if "PYTEST_CURRENT_TEST" in os.environ:
        print('DBG> Running inside a pytest -> not showing error message.')
        return
    info = QMessageBox(parent=parent)
    info.setIcon(QMessageBox.Icon.Warning)
    info.setText('The "_shelx_hkl_checksum" is not\nconsistent with the .hkl file content!')
    info.setModal(True)
    info.show()


def show_general_warning(parent, warn_text: str = '', info_text: str = '', window_title=' ') -> None:
    """
    A message box to display if the checksums do not agree.
    warn_text is displayed bold.
    info_text is displayed regular.
    """
    if not warn_text:
        return None
    if "PYTEST_CURRENT_TEST" in os.environ:
        print(f'DBG> Running inside a pytest -> not showing error message:\n{warn_text}\n{info_text}')
        return None
    box = QMessageBox(parent=parent)
    box.setTextFormat(QtCore.Qt.TextFormat.AutoText)
    box.setWindowTitle(window_title)
    box.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextBrowserInteraction)
    box.setText(warn_text)
    box.setModal(True)
    if info_text:
        box.setInformativeText(info_text)
        box.setStyleSheet("QLabel{min-width:600 px; font-size: 14px;}")
    box.show()
    if parent is None:
        box.exec()
        box.close()


def show_keyword_help(parent, helptext: str, title: str = ''):
    """
    A window to display help texts from the CIF dictionaries.
    """
    nlines = len(helptext.splitlines())
    window = QMainWindow(parent=parent)
    window.setWindowTitle(title)

    def close_window(event) -> None:
        if event.key() == QtCore.Qt.Key.Key_Escape:
            window.close()

    window.keyPressEvent = close_window
    widget = QFrame()
    layout = QVBoxLayout()
    button = QPushButton('close')
    textedit = QTextEdit()
    textedit.setReadOnly(True)
    textedit.setFontFamily('monospace')
    textedit.setText(helptext)
    layout.addWidget(textedit)
    layout.addWidget(button)
    widget.setLayout(layout)
    window.setCentralWidget(widget)
    width = textedit.fontMetrics().horizontalAdvance('X' * 70)
    height = textedit.fontMetrics().horizontalAdvance('X' * nlines)
    textedit.setMinimumWidth(max([600, width]))
    textedit.setMinimumHeight(max([400, height]))
    window.move(300, 100)
    window.show()
    button.clicked.connect(window.close)


def show_ok_cancel_warning(parent, warn_text: str = '') -> bool:
    box = QMessageBox(parent=parent)
    box.setTextFormat(QtCore.Qt.TextFormat.AutoText)
    box.setWindowTitle(" ")
    box.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextBrowserInteraction)
    box.setText(warn_text)
    box.setStandardButtons(QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
    box.setDefaultButton(QMessageBox.StandardButton.Ok)
    box.setModal(True)
    box.show()
    return box.result() == QMessageBox.StandardButton.Ok


def show_update_warning(parent, remote_version: int = 0) -> None:
    """
    A message box to display if the checksums do not agree.
    """
    warn_text = "A newer version {} of FinalCif is available under: <br>" \
                "<a href='https://dkratzert.de/finalcif.html'>" \
                "https://dkratzert.de/finalcif.html</a>"
    box = QMessageBox(parent)
    box.setTextFormat(QtCore.Qt.TextFormat.AutoText)
    box.setWindowTitle(" ")
    box.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextBrowserInteraction)
    if sys.platform.startswith("win"):
        warn_text += r"<br><br>Updating now will end all running FinalCIF programs!"
        update_button = box.addButton('Update Now', QMessageBox.ButtonRole.AcceptRole)
        update_button.clicked.connect(lambda: do_update_program(str(remote_version)))
    box.setText(warn_text.format(remote_version))
    box.setModal(True)
    box.show()


def bad_z_message(parent, z: float) -> None:
    zinfo = QMessageBox(parent)
    zinfo.setIcon(QMessageBox.Icon.Information)
    zinfo.setText(f'The number of formula units Z={z:.0f} is probably wrong.\n'
                  f'You may restart refinement with a correct value.')
    zinfo.setModal(True)
    zinfo.show()


def show_bug_found_warning(logfile) -> None:
    window = QMainWindow()
    title = 'Congratulations, you found a bug in FinalCif!'
    text = (f'<br>Please send the file <br>'
            f'<a href={logfile.resolve()}>{logfile.resolve()}</a> '
            f'<br>to Daniel Kratzert:  '
            f'<a href="mailto:dkratzert@gmx.de?subject=FinalCif version {VERSION} crash report">'
            f'dkratzert@gmx.de</a><br>'
            f'<br>If possible, the corresponding CIF file is also desired.')
    box = QMessageBox(parent=window)
    box.setWindowTitle('Warning')
    box.setText(title)
    box.setInformativeText(text)
    box.setTextFormat(QtCore.Qt.TextFormat.RichText)
    box.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextBrowserInteraction)
    box.exec()
    window.show()


def show_yes_now_question(title: str, question: str, parent=None) -> bool:
    response = QMessageBox.question(parent, title, question,
                                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                                    QMessageBox.StandardButton.No)
    if response == QMessageBox.StandardButton.Yes:
        return True
    else:
        return False


def cif_file_open_dialog(parent: object = None, filter: str = "CIF file (*.cif)", last_dir='', options=None) -> str:
    """
    Returns a cif file name from a file dialog.
    """
    filename, _ = compat.getopenfilename(parent=parent,
                                         caption='Open a .cif File',
                                         basedir=last_dir,
                                         filters=filter,
                                         selectedfilter=filter,
                                         options=options
                                         )
    return filename


def cif_file_save_dialog(filename: str, parent=None) -> str:
    """
    Returns a cif file name from a file dialog.
    """
    filter = "CIF file (*.cif)"
    filename, _ = compat.getsavefilename(parent=parent,
                                         filters=filter,
                                         caption='Save .cif File',
                                         selectedfilter=filter)
    return filename


def video_file_open_dialog(parent: object = None, filter: str = "Video file (*.vzs; *.jpg)", last_dir='', options=None) -> str:
    filename, _ = compat.getopenfilename(parent=parent,
                                         caption='Open a crystal video file',
                                         basedir=last_dir,
                                         filters=filter,
                                         selectedfilter=filter,
                                         options=options
                                         )
    return filename


if __name__ == '__main__':
    from qtpy import QtWidgets
    from qtpy.QtWidgets import QApplication

    app = QApplication.instance()
    if app is None:
        app = QApplication([])

    window = QtWidgets.QMainWindow()
    w = QtWidgets.QWidget()
    window.setCentralWidget(w)
    l = QtWidgets.QVBoxLayout()
    w.setLayout(l)
    # answer = show_yes_now_question(title='Delete templates', question='Fobar?', parent=w)
    # bad_z_message(parent=w, z=3.0)
    # show_update_warning(parent=w, remote_version=123)
    # show_bug_found_warning(Path(r'test.txt'))
    # show_ok_cancel_warning(parent=w, warn_text='foobar')
    # show_keyword_help(parent=w, helptext="This is a helptext", title='A Title')
    # show_general_warning(parent=w, warn_text='Warning text', info_text='Info text', window_title='Title')
    # show_hkl_checksum_warning(parent=w)
    # show_res_checksum_warning(parent=w)
    # unable_to_open_message(parent=w, not_ok=Exception('foo'), filepath=Path('C:/foo.txt'))
    # do_update_program('127')
    # window.show()

    app.exec()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'mainwindow.ui'
##
##
################################################################################

from qtpy.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from qtpy.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from qtpy.QtWidgets import (QAbstractItemView, QApplication, QGridLayout, QGroupBox,
    QListWidget, QListWidgetItem, QMainWindow, QPushButton,
    QSizePolicy, QSpacerItem, QSplitter, QTextEdit,
    QVBoxLayout, QWidget)

class Ui_MultitableWindow(object):
    def setupUi(self, MultitableWindow):
        if not MultitableWindow.objectName():
            MultitableWindow.setObjectName(u"MultitableWindow")
        MultitableWindow.resize(809, 588)
        self.Mainwidget = QWidget(MultitableWindow)
        self.Mainwidget.setObjectName(u"Mainwidget")
        self.gridLayout = QGridLayout(self.Mainwidget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.cif_files_button = QPushButton(self.Mainwidget)
        self.cif_files_button.setObjectName(u"cif_files_button")

        self.gridLayout.addWidget(self.cif_files_button, 0, 0, 1, 1)

        self.report_button = QPushButton(self.Mainwidget)
        self.report_button.setObjectName(u"report_button")

        self.gridLayout.addWidget(self.report_button, 0, 1, 1, 2)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer, 0, 3, 1, 1)

        self.removeButton = QPushButton(self.Mainwidget)
        self.removeButton.setObjectName(u"removeButton")

        self.gridLayout.addWidget(self.removeButton, 0, 4, 1, 1)

        self.splitter_2 = QSplitter(self.Mainwidget)
        self.splitter_2.setObjectName(u"splitter_2")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.splitter_2.sizePolicy().hasHeightForWidth())
        self.splitter_2.setSizePolicy(sizePolicy)
        self.splitter_2.setOrientation(Qt.Vertical)
        self.groupBox = QGroupBox(self.splitter_2)
        self.groupBox.setObjectName(u"groupBox")
        self.verticalLayout = QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.CifFileListListWidget = QListWidget(self.groupBox)
        self.CifFileListListWidget.setObjectName(u"CifFileListListWidget")
        self.CifFileListListWidget.setDragDropMode(QAbstractItemView.DragDrop)
        self.CifFileListListWidget.setDefaultDropAction(Qt.CopyAction)
        self.CifFileListListWidget.setAlternatingRowColors(True)

        self.verticalLayout.addWidget(self.CifFileListListWidget)

        self.splitter_2.addWidget(self.groupBox)
        self.groupBox_2 = QGroupBox(self.splitter_2)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.verticalLayout_2 = QVBoxLayout(self.groupBox_2)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.OutputTextEdit = QTextEdit(self.groupBox_2)
        self.OutputTextEdit.setObjectName(u"OutputTextEdit")
        sizePolicy.setHeightForWidth(self.OutputTextEdit.sizePolicy().hasHeightForWidth())
        self.OutputTextEdit.setSizePolicy(sizePolicy)
        self.OutputTextEdit.setMinimumSize(QSize(0, 0))
        self.OutputTextEdit.setReadOnly(True)
        self.OutputTextEdit.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByMouse)

        self.verticalLayout_2.addWidget(self.OutputTextEdit)

        self.splitter_2.addWidget(self.groupBox_2)

        self.gridLayout.addWidget(self.splitter_2, 1, 0, 2, 5)

        MultitableWindow.setCentralWidget(self.Mainwidget)

        self.retranslateUi(MultitableWindow)
    # setupUi

    def retranslateUi(self, MultitableWindow):
        MultitableWindow.setWindowTitle(QCoreApplication.translate("MultitableWindow", u"Mutitable", None))
        self.cif_files_button.setText(QCoreApplication.translate("MultitableWindow", u"Select CIF files", None))
        self.report_button.setText(QCoreApplication.translate("MultitableWindow", u"Generate Report", None))
#if QT_CONFIG(tooltip)
        self.removeButton.setToolTip(QCoreApplication.translate("MultitableWindow", u"<html><head/><body><p>Removes the currently selected file from the list.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.removeButton.setText(QCoreApplication.translate("MultitableWindow", u"Remove Current File", None))
        self.groupBox.setTitle(QCoreApplication.translate("MultitableWindow", u"Files List", None))
        self.groupBox_2.setTitle(QCoreApplication.translate("MultitableWindow", u"Program Output", None))
    # retranslateUi




import os

from qtpy.QtWidgets import QApplication
app = QApplication.instance()
if app is None:
    app = QApplication([])

os.environ["RUNNING_TEST"] = 'True'
import shutil
import unittest
from pathlib import Path

from finalcif.cif.cif_file_io import CifContainer
from finalcif.gui.import_selector import ImportSelector
from finalcif.tools.settings import FinalCifSettings

data = Path('tests')
testdata = Path('test-data')




class MyTestCase(unittest.TestCase):
    def setUp(self) -> None:
        imp_cif = CifContainer(data / 'statics/import_cif.cif')
        shutil.copyfile(testdata / 'p21c.cif', testdata / 'p21c-copy.cif')
        targetcif = CifContainer(testdata / 'p21c-copy.cif')
        settings = FinalCifSettings()
        self.imp = ImportSelector(None, import_cif=imp_cif, target_cif=targetcif, settings=settings)
        self.imp._empty_saved_selection()
        self.imp.show_import_window()

    def tearDown(self) -> None:
        Path(testdata / 'p21c-copy.cif').unlink(missing_ok=True)
        self.imp.deleteLater()
        self.imp.close()

    def test_keys_to_import(self):
        self.assertEqual(3, self.imp.keys_to_import)
        self.assertEqual(1, self.imp.loops_to_import)

    def test_import_methods(self):
        self.assertEqual(['_foo_bar', '_hello'], self.imp.get_keys(include=True))
        self.assertEqual([['_my_atom_type_symbol',
                           '_my_atom_type_description',
                           '_my_atom_type_scat_dispersion_real']], self.imp.get_loops(include=True))

    def test_import_methods_excluded(self):
        self.assertEqual(['_cell_length_a'], self.imp.get_keys(include=False))
        self.assertEqual([], self.imp.get_loops(include=False))

    def test_other(self):
        self.assertEqual('import_cif.cif', self.imp.import_cif.filename)
        self.assertEqual('p21c-copy.cif', self.imp.target_cif.filename)


if __name__ == '__main__':
    unittest.main()



import unittest

import gemmi
from packaging import version


class MyGemmiTestsForAPI(unittest.TestCase):
    def setUp(self) -> None:
        self.doc: gemmi.cif.Document = gemmi.cif.Document()
        self.doc.add_new_block('block1')
        self.doc.add_new_block('block2')

    def test_blocks_in_doc(self):
        self.assertEqual('<gemmi.cif.Block block1>', str(self.doc.find_block('block1')))  # add assertion here
        self.assertEqual('[<gemmi.cif.Block block1>, <gemmi.cif.Block block2>]', str(list(self.doc)))  # add assertion here
        self.assertTrue('block2' in [x.name for x in self.doc])
        if version.parse(gemmi.__version__) > version.parse('0.5.7'):
            self.assertTrue('block2' in self.doc)

    def test_delete_block(self):
        self.assertEqual(2, len(self.doc))
        del self.doc[1]
        self.assertEqual(1, len(self.doc))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum

from qtpy import QtCore
from qtpy.QtCore import QAbstractTableModel, QModelIndex, Qt
from qtpy.QtGui import QColor, QBrush

from finalcif.cif.text import retranslate_delimiter

white = QColor(255, 255, 255)
light_green = QColor(217, 255, 201)
light_blue = QColor(244, 244, 249)
blue = QColor(102, 150, 179)
yellow = QColor(250, 247, 150)
light_red = QColor(254, 191, 189)

COLUMN_HEADERS = ['CIF Value', 'From Data Source', 'Own Data']


class Column(IntEnum):
    CIF = 0
    DATA = 1
    EDIT = 2


@dataclass
class CifRowData:
    """Data for a single row in the CIF table.

    The ``raw_value`` and ``strval`` fields are transient: they carry the
    original CIF value through the population loop in
    ``AppWindow.fill_cif_table`` so that a single list of ``CifRowData``
    objects can replace both the model rows *and* the former ``prepared``
    dicts.  The model itself never reads them.
    """
    key: str = ''
    cif_value: str = ''
    data_value: str = ''
    edit_value: str = ''
    cif_color: QColor | None = None
    data_color: QColor | None = None
    edit_color: QColor | None = None
    is_separator: bool = False
    # Transient fields used only during table population:
    raw_value: str = ''
    strval: str = ''


class CifTableModel(QAbstractTableModel):
    """Model for the main CIF table holding key-value-pair data."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self._rows: list[CifRowData] = []
        self._key_index: dict[str, int] = {}  # key → row for O(1) lookup

    # --- Core QAbstractTableModel interface ---

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self._rows)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return 3

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        row_data = self._rows[index.row()]
        col = index.column()
        if row_data.is_separator:
            if role == Qt.ItemDataRole.BackgroundRole:
                brush = QBrush(blue)
                brush.setStyle(Qt.BrushStyle.DiagCrossPattern)
                return brush
            if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
                return ''
            return None
        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            if col == 0:
                return row_data.cif_value
            elif col == 1:
                return row_data.data_value
            elif col == 2:
                return row_data.edit_value
        if role == Qt.ItemDataRole.BackgroundRole:
            if col == 0:
                return row_data.cif_color
            elif col == 1:
                return row_data.data_color
            elif col == 2:
                return row_data.edit_color
        return None

    def setData(self, index: QModelIndex, value, role: int = Qt.ItemDataRole.EditRole) -> bool:
        if not index.isValid():
            return False
        row_data = self._rows[index.row()]
        col = index.column()
        if role == Qt.ItemDataRole.EditRole:
            if col == 0:
                row_data.cif_value = value
            elif col == 1:
                row_data.data_value = value
            elif col == 2:
                row_data.edit_value = value
            self.dataChanged.emit(index, index, [role])
            return True
        if role == Qt.ItemDataRole.BackgroundRole:
            if col == 0:
                row_data.cif_color = value
            elif col == 1:
                row_data.data_color = value
            elif col == 2:
                row_data.edit_color = value
            self.dataChanged.emit(index, index, [role])
            return True
        return False

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        row_data = self._rows[index.row()]
        if row_data.is_separator:
            return Qt.ItemFlag.ItemIsEnabled
        base = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable
        if index.column() == 2:
            base |= Qt.ItemFlag.ItemIsEditable
        return base

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                if 0 <= section < len(COLUMN_HEADERS):
                    return COLUMN_HEADERS[section]
            elif orientation == Qt.Orientation.Vertical:
                if 0 <= section < len(self._rows):
                    return self._rows[section].key
        if role == Qt.ItemDataRole.BackgroundRole and orientation == Qt.Orientation.Vertical:
            if 0 <= section < len(self._rows) and self._rows[section].is_separator:
                brush = QBrush(blue)
                brush.setStyle(Qt.BrushStyle.DiagCrossPattern)
                return brush
        return None

    # --- Helper methods ---

    @property
    def vheaderitems(self) -> list[str]:
        """Return list of CIF keywords in row order (compatible with old MyCifTable.vheaderitems)."""
        return [r.key for r in self._rows]

    def row_from_key(self, key: str) -> int:
        """Return row index for a given CIF keyword (O(1) via index cache)."""
        try:
            return self._key_index[key]
        except KeyError:
            raise ValueError(f'Key {key!r} not found in model')

    def has_key(self, key: str) -> bool:
        return key in self._key_index

    def add_row(self, key: str, cif_value: str = '', data_value: str = '',
                edit_value: str = '', position: int | None = None,
                is_separator: bool = False) -> int:
        """Insert a new row. Returns the row index where it was inserted."""
        if position is None:
            position = len(self._rows)
        row_data = CifRowData(
            key=key,
            cif_value=cif_value,
            data_value=data_value,
            edit_value=edit_value,
            is_separator=is_separator,
        )
        self.beginInsertRows(QModelIndex(), position, position)
        self._rows.insert(position, row_data)
        # Rebuild key index for rows at and after the insertion point.
        for i in range(position, len(self._rows)):
            self._key_index[self._rows[i].key] = i
        self.endInsertRows()
        return position

    def bulk_load(self, rows: list[CifRowData]) -> None:
        """Replace all rows at once — much faster than per-row add_row.

        Uses a single ``beginResetModel``/``endResetModel`` pair instead
        of N ``beginInsertRows``/``endInsertRows`` calls.
        """
        self.beginResetModel()
        self._rows = list(rows)
        self._key_index = {r.key: i for i, r in enumerate(self._rows)}
        self.endResetModel()

    def remove_row(self, row: int) -> str:
        """Remove a row by index. Returns the key of the removed row."""
        if 0 <= row < len(self._rows):
            key = self._rows[row].key
            self.beginRemoveRows(QModelIndex(), row, row)
            del self._rows[row]
            # Rebuild index for shifted rows.
            self._key_index.pop(key, None)
            for i in range(row, len(self._rows)):
                self._key_index[self._rows[i].key] = i
            self.endRemoveRows()
            return key
        return ''

    def get_row_data(self, row: int) -> CifRowData | None:
        if 0 <= row < len(self._rows):
            return self._rows[row]
        return None

    def set_cell_text(self, row: int, column: int, text: str) -> None:
        """Set text for a specific cell."""
        if 0 <= row < len(self._rows):
            idx = self.index(row, column)
            self.setData(idx, text, Qt.ItemDataRole.EditRole)

    def set_cell_color(self, row: int, column: int, color: QColor | None) -> None:
        """Set background color for a specific cell."""
        if 0 <= row < len(self._rows):
            idx = self.index(row, column)
            self.setData(idx, color, Qt.ItemDataRole.BackgroundRole)

    def get_cell_text(self, row: int, column: int) -> str:
        """Get text for a specific cell."""
        idx = self.index(row, column)
        return self.data(idx, Qt.ItemDataRole.DisplayRole) or ''

    def clear_all(self) -> None:
        """Remove all rows."""
        if self._rows:
            self.beginResetModel()
            self._rows.clear()
            self._key_index.clear()
            self.endResetModel()

    def set_separator(self, row: int) -> None:
        """Mark a row as the separator line."""
        if 0 <= row < len(self._rows):
            self._rows[row].is_separator = True
            self.dataChanged.emit(
                self.index(row, 0), self.index(row, 2),
                [Qt.ItemDataRole.BackgroundRole, Qt.ItemDataRole.DisplayRole],
            )



#   ----------------------------------------------------------------------------
#   "THE BEER-WARE LICENSE" (Revision 42):
#   Daniel Kratzert <dkratzert@gmx.de> wrote this file.  As long as you retain
#   this notice you can do whatever you want with this stuff. If we meet some day,
#   and you think this stuff is worth it, you can buy me a beer in return.
#   ----------------------------------------------------------------------------
import sys
from shutil import which
import unittest

addr = """Albert-Ludwigs-Universität Freiburg
Institut für Anorganische und Analytische Chemie
Albertstraße 21
Freiburg i. Br.
79104
Germany"""


def get_platon_exe() -> str:
    if sys.platform.startswith('win'):
        platon_exe = r'C:\pwt\platon.exe'
    else:
        platon_exe = which('platon')
    return platon_exe


def processevents() -> None:
    from qtpy.QtWidgets import QApplication
    app = QApplication.instance()
    if app is not None:
        app.processEvents()


class AppWindowTestCase(unittest.TestCase):
    def setUp(self) -> None:
        import gc
        gc.collect()
        super().setUp()

    def tearDown(self) -> None:
        if hasattr(self, 'app') and getattr(self, 'app') is not None:
            try:
                self.app.deleteLater()
                self.app.close()
            except RuntimeError:
                pass
            self.app = None
        super().tearDown()



import unittest

from finalcif.tools.sumformula import formula_str_to_dict, sum_formula_to_html


class MyTestCase(unittest.TestCase):

    def test_sumform1(self):
        self.assertEqual({'S': 1.0, 'Sn': 1.0}, formula_str_to_dict("SSn"))

    def test_sumform2(self):
        self.assertEqual({'Cl': 1.0, 'S': 1.0}, formula_str_to_dict("S1Cl"))

    def test_sumform3(self):
        self.assertEqual({'C': 12.0, 'H': 6.0, 'Mn': 7.0, 'O': 3.0}, formula_str_to_dict("C12H6O3Mn7"))

    def test_sumform4(self):
        self.assertEqual({'C': 12.0, 'H': 60.0, 'Mn': 7.0, 'O': 3.0}, formula_str_to_dict("C12 H60 O3 Mn7"))

    def test_sumform5(self):
        self.assertEqual({'C': 12.0, 'H': 60.0, 'Mn': 7.0, 'O': 3.0}, formula_str_to_dict("C12 H60 O3  Mn 7"))

    def test_sumform6(self):
        self.assertEqual({'C': 13.0, 'Cs': 12.0, 'H': 60.0, 'Mn': 7.0, 'O': 3.0},
                         formula_str_to_dict("C13Cs12 H60 O3  Mn 7"))

    def test_sumform7(self):
        self.assertEqual({'C': 1.0, 'H': 1.0, 'Mn': 1.0}, formula_str_to_dict("CHMn\n"))

    def test_sumform8(self):
        self.assertEqual({'Hallo': 1.0}, formula_str_to_dict("Hallo"))

    def test_sumform9(self):
        self.assertEqual({'+': 1.0, 'H': 3.0, 'O': 1.0}, formula_str_to_dict("H3O+"))

    def test_sumform10(self):
        self.assertEqual({'Al': 0.12, 'C': 4.0, 'F': 4.36, 'H': 2.91, 'Ni': 0.12, 'O': 0.48},
                         formula_str_to_dict('C4 H2.91 Al0.12 F4.36 Ni0.12 O0.48'))

    def test_sumform11(self):
        self.assertEqual({'C': 4.0, 'H': 8.0, 'O': 2.0}, formula_str_to_dict('C4H6O*5(H2O)'))

    def test_sumform12(self):
        self.assertEqual({'B': 1.0, 'C': 15.0, 'F': 2.0, 'H': 23.0, 'N': 2.0, 'O': 1.0, 'Si': 2.0},
                         formula_str_to_dict('C15 H23 B F2 N2 O Si2'))

    def test_sumform13(self):
        self.assertEqual({'B': 1.0, 'C': 15.0, 'F': 2.0, 'H': 23.0, 'N': 2.0, 'O': 1.0, 'Si': 1.0},
                         formula_str_to_dict('C15 H23 B F2 N2 O Si'))

    def test_sumform14(self):
        self.assertEqual({'B': 1.0, 'C': 15.0, 'F': 2.0, 'H': 23.0, 'I': 1.0, 'N': 2.0, 'Os': 1.0},
                         formula_str_to_dict('C15 H23 B F2 N2 Os I'))


class TestSumformHTLM(unittest.TestCase):

    def test_sumform_to_html(self):
        self.assertEqual('<html><body>C<sub>12</sub>H<sub>6</sub>O<sub>3</sub>Mn<sub>7</sub></body></html>',
                         sum_formula_to_html({'C': 12, 'H': 6, 'O': 3, 'Mn': 7}))

    def test_sumform_to_html_with_difficult_elements(self):
        self.assertEqual(
            '<html><body>C<sub>15</sub>H<sub>23</sub>BF<sub>2</sub>N<sub>2</sub>OSi<sub>2</sub></body></html>',
            sum_formula_to_html(formula_str_to_dict('C15H23BF2N2OSi2')))


if __name__ == '__main__':
    unittest.main()



from docxtpl import RichText
from lxml import etree

from finalcif.app_path import application_path


def xml_to_html(xml_string: str | RichText):
    """
    Transforms XML to HTML using XSLT.
    """
    xslt_file = application_path / 'template/xsl/xmltohtml.xsl'
    xml_string = (fr'''
    <w:document xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">
        {xml_string}
    </w:document>
    ''')

    with open(xslt_file, 'rb') as f:
        xslt_data = f.read()

    xml_doc = etree.fromstring(xml_string)
    xslt_doc = etree.fromstring(xslt_data)

    transform = etree.XSLT(xslt_doc)
    result = str(transform(xml_doc)).splitlines()
    return ''.join(result[1:-1]).strip()


if __name__ == '__main__':
    xml_string = (
        '<w:r><w:t xml:space="preserve">All hydrogen atoms were refined isotropic </w:t></w:r><w:r><w:t '
        'xml:space="preserve">on calculated positions using a riding model with their </w:t></w:r><w:r><w:rPr>'
        '<w:i/></w:rPr><w:t xml:space="preserve">U</w:t></w:r><w:r><w:rPr><w:vertAlign w:val="subscript"/>'
        '</w:rPr><w:t xml:space="preserve">iso</w:t></w:r><w:r><w:t xml:space="preserve"> values constrained to '
        '1.5 times the </w:t></w:r><w:r><w:rPr><w:i/></w:rPr><w:t xml:space="preserve">U</w:t></w:r><w:r><w:rPr>'
        '<w:vertAlign w:val="subscript"/></w:rPr><w:t xml:space="preserve">eq</w:t></w:r><w:r><w:t '
        'xml:space="preserve"> of their pivot atoms for terminal sp</w:t></w:r><w:r><w:rPr><w:vertAlign '
        'w:val="superscript"/></w:rPr><w:t xml:space="preserve">3</w:t></w:r><w:r><w:t xml:space="preserve"> '
        'carbon atoms and 1.2 times for all other carbon atoms.</w:t></w:r>')
    print(xml_to_html(xml_string))



from enum import IntEnum
from typing import TYPE_CHECKING

from qtpy import QtCore, QtGui
from qtpy.QtCore import Signal, Qt, QObject, QEvent, QSize
from qtpy.QtGui import QTextOption, QFontMetrics, QContextMenuEvent, QFont, QColor
from qtpy.QtWidgets import QPlainTextEdit, QFrame, QAbstractScrollArea, QWidget

from finalcif.gui.edit_button import FloatingButtonWidget
from finalcif.gui.new_key_dialog import NewKey
from finalcif.gui.validators import validators

if TYPE_CHECKING:
    from finalcif.gui.cif_table_view import CifTableView

light_red = QColor(254, 191, 189)


class Column(IntEnum):
    CIF = 0
    DATA = 1
    EDIT = 2


class MyQPlainTextEdit(QPlainTextEdit):
    """
    A special plaintextedit with convenient methods to set the background color and other things.
    """
    templateRequested = Signal(int)
    new_key = Signal(str)
    to_be_shortened = {'_shelx_hkl_file', '_shelx_res_file', '_shelx_fab_file', '_shelx_fcf_file',
                       '_iucr_refine_instructions_details', '_iucr_refine_fcf_details'}

    def __init__(self, parent=None, *args):
        """
        Plaintext edit field for most of the table cells.
        """
        super().__init__(parent, *args)
        self.color = None
        self.cif_key = ''
        self.edit_button = None
        self.default_palette = self.palette()
        # self.increse_font_size()
        self.parent: CifTableView = parent
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setTabChangesFocus(True)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setWordWrapMode(QTextOption.WrapMode.WordWrap)
        self.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.textChanged.connect(self._on_text_changed)

    def _on_text_changed(self) -> None:
        # Skip expensive indexAt lookup while the table is being bulk-populated
        # (updates are disabled).  The caller will resize all rows at the end.
        if self.parent.updatesEnabled():
            self.parent.resizeRowToContents(self.row)

    def increse_font_size(self):
        font = QFont()
        font.setPointSize(self.document().defaultFont().pointSize() + 1)
        self.setFont(font)

    def __str__(self):
        return self.toPlainText()

    @property
    def fontmetric(self):
        return QFontMetrics(self.document().defaultFont())

    def contextMenuEvent(self, event: QContextMenuEvent):
        menu = self.createStandardContextMenu(event.pos())
        action_copy_vhead = menu.addAction("Copy CIF Keyword")
        deleterow = menu.addAction("Delete Row")
        menu.addSeparator()
        action_template = menu.addAction("Text Template")
        new_key = menu.addAction('Add new CIF keys')
        action_copy_vhead.triggered.connect(self.copy_vhead_item)
        action_template.triggered.connect(self._on_create_template)
        deleterow.triggered.connect(self._delete_row)
        new_key.triggered.connect(self._add_cif_keys)
        menu.exec(event.globalPos())

    def _add_cif_keys(self) -> None:
        new_key = NewKey(self)
        new_key.show()
        new_key.new_key_added.connect(lambda x: self.new_key.emit(x))

    def _on_create_template(self) -> None:
        self.templateRequested.emit(self.row)

    def _delete_row(self) -> None:
        self.parent.delete_row(self.row)

    def copy_vhead_item(self, row):
        """
        Copies the CIF key of the row to clipboard.
        """
        from qtpy import QtWidgets
        if self.cif_key:
            clipboard = QtWidgets.QApplication.clipboard()
            clipboard.setText(self.cif_key)

    @property
    def row(self) -> int:
        return self.parent.indexAt(self.pos()).row()

    @property
    def column(self) -> int:
        return self.parent.indexAt(self.pos()).column()

    def setBackground(self, color: QColor) -> None:
        """
        Set background color of the text field.
        """
        # self.setStyleSheet("background-color: {};".format(str(color.name())))
        palette = self.palette()
        palette.setColor(QtGui.QPalette.ColorRole.Base, color)
        #palette.setColor(QtGui.QPalette.ColorRole.AlternateBase, color)
        self.setPalette(palette)
        self.default_palette = palette

    def getBackgroundColor(self) -> QtGui.QColor:
        palette = self.palette()
        background_color = palette.color(QtGui.QPalette.ColorRole.Base)
        return background_color

    def setUneditable(self):
        self.setReadOnly(True)

    MAX_DISPLAY_LINES = 5

    def setText(self, text: str, color: QColor = None, column: int | None = None):
        """
        Set text of a Plaintextfield with lines wrapped at newline characters.
        """
        if color:
            self.setBackground(color)
        if not text:
            return
        if column == Column.CIF and self.cif_key in self.to_be_shortened:
            lines = text.splitlines(keepends=True)
            if len(lines) > self.MAX_DISPLAY_LINES:
                text = ''.join(lines[:self.MAX_DISPLAY_LINES]).rstrip() + '\n[...]'
        # Block signals to prevent textChanged → resizeRowToContents per call.
        # The caller (CifTableView.setText) handles resizing explicitly.
        was_blocked = self.signalsBlocked()
        self.blockSignals(True)
        self.setPlainText(text)
        self.blockSignals(was_blocked)

    def eventFilter(self, widget: QWidget, event: QEvent):
        """
        Event filter to ignore wheel events in comboboxes to prevent accidental changes to them.
        """
        if event.type() == QEvent.Type.Wheel and widget and not widget.hasFocus():
            event.ignore()
            return True
        return QObject.eventFilter(self, widget, event)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        if self.edit_button and self.column == Column.EDIT:
            self.edit_button.update_position()

    def enterEvent(self, a0):
        super().enterEvent(a0)
        if self.column == Column.EDIT:
            if not self.edit_button:
                self.edit_button = FloatingButtonWidget(parent=self)
                self.edit_button.floatingButtonClicked.connect(self._on_create_template)
                self.edit_button.update_position()
            self.edit_button.show()

    def event(self, e: QtCore.QEvent):
        if e.type() == QtCore.QEvent.Type.InputMethodQuery:
            # Skip expensive indexAt lookup while the table is being bulk-populated.
            if self.parent.updatesEnabled():
                self.parent.setCurrentCell(self.row, self.column)
        return super().event(e)

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        super().keyPressEvent(event)
        if self.column == Column.EDIT:
            self.validate_text(self.toPlainText())

    def validate_text(self, text: str):
        validator = validators.get(self.cif_key, None)
        if validator and not validator.valid(text):
            self.setBadStyle()
            self.setToolTip(validator.help_text)
        else:
            self.setToolTip('')
            self.setRegularStyle()

    def setBadStyle(self) -> None:
        palette = self.palette()
        palette.setColor(QtGui.QPalette.ColorRole.Base, light_red)
        self.setPalette(palette)

    def setRegularStyle(self) -> None:
        self.setPalette(self.default_palette)

    def leaveEvent(self, a0: QEvent) -> None:
        super().leaveEvent(a0)
        if self.edit_button and self.column == Column.EDIT:
            self.edit_button.hide()

    def getText(self):
        return self.toPlainText()

    def sizeHint(self) -> QSize:
        """Text field sizes are scaled to text length"""
        max_size = QSize(100, 350)
        if len(self.toPlainText()) > 500:
            return max_size
        rect = self.fontmetric.boundingRect(self.contentsRect(), Qt.TextFlag.TextWordWrap, self.toPlainText())
        size = QSize(100, rect.height() + 14)
        if size.height() > 50:
            size = QSize(100, rect.height() + 24)
        if size.height() > 350:
            # Prevent extreme height for long text:
            size = max_size
        return size


class PlainTextEditTemplate(QPlainTextEdit):
    """
    A special plaintextedit for equipment and properties.
    """
    templateRequested = Signal(int)

    def __init__(self, parent=None, *args, **kwargs):
        """
        Plaintext edit field for most of the table cells.
        :param parent:
        :param minheight: minimum height of the widget.
        """
        super().__init__(parent, *args, **kwargs)
        self.cif_key = ''
        self.parent = parent
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setTabChangesFocus(True)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setWordWrapMode(QTextOption.WrapMode.WordWrap)
        self.fontmetric = QFontMetrics(self.document().defaultFont())
        self.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.textChanged.connect(lambda: self.parent.resizeRowsToContents())

    def setText(self, text: str, color=None):
        """
        Set text of a Plaintextfield with lines wrapped at newline characters.
        """
        if color:
            self.setBackground(color)
        self.setPlainText(text)

    def eventFilter(self, widget: QWidget, event: QEvent):
        """
        Event filter to ignore wheel events in comboboxes to prevent accidental changes to them.
        """
        if event.type() == QEvent.Type.Wheel and widget and not widget.hasFocus():
            event.ignore()
            return True
        return QObject.eventFilter(self, widget, event)

    def getText(self):
        return self.toPlainText()

    def sizeHint(self) -> QSize:
        """Text field sizes are scaled to text length"""
        rect = self.fontmetric.boundingRect(self.contentsRect(), Qt.TextFlag.TextWordWrap, self.toPlainText())
        size = QSize(100, rect.height() + 14)
        if size.height() > 300:
            # Prevent extreme height for long text:
            size = QSize(100, 300)
        return size


if __name__ == '__main__':
    import sys
    import random
    from qtpy.QtWidgets import QApplication, QTableWidget

    app = QApplication(sys.argv)
    window = QTableWidget()
    window.setColumnCount(3)
    window.setRowCount(10)
    # stretch the last table colum:
    window.horizontalHeader().setStretchLastSection(True)
    window.setHorizontalHeaderLabels(['CIF', 'Data', 'Edit'])
    for row in range(10):
        for col in range(3):
            window.setCellWidget(row, col, MyQPlainTextEdit(window))
            w = window.cellWidget(row, col)
            if col == 2:
                w.setText(f'Hello World {random.randint(0, 10000)} {random.randint(0, 10000)}')
            w.setMinimumHeight(50)
            w.setMinimumWidth(150)
    window.resizeRowsToContents()
    window.resizeColumnsToContents()
    window.show()
    window.setMinimumSize(600, 400)

    sys.exit(app.exec())



"""
This file combines all major CIF dictionarys in order to check for the validity of CIF keys
and to display help texts.
"""

from finalcif.cif.core_dict import cif_core
from finalcif.cif.modulation_dict import modulation_dict
from finalcif.cif.powder_dict import powder_dict
from finalcif.cif.restraints_dict import restraints_dict
from finalcif.cif.twin_dict import twinning_dict

additional_keywords = {
    '_exptl_absorpt_special_details'          : '<pre><h2>_exptl_absorpt_special_details</h2>'
                                                "Details of the absorption correction process applied to the"
                                                "measured intensities that cannot otherwise be given using"
                                                "other data items from the EXPTL_ABSORBT category."
                                                '</pre>'
                                                '<br><p><h4>Type:</h4> Text</p>\n',
    '_shelx_estimated_absorpt_T_max'          : '<pre><h2>_shelx_estimated_absorpt_T_max</h2>'
                                                'The maximum and minimum transmission factors estimatedby SHELXL.</pre>'
                                                '<br><p><h4>Type:</h4> number (int or float)</p>\n'
                                                '<br><p><h4>Limits:</h4> 0.0:1.0 </p>',
    '_shelx_estimated_absorpt_T_min'          : '<pre><h2>_shelx_estimated_absorpt_T_min</h2>'
                                                'The maximum and minimum transmission factors estimatedby SHELXL.</pre>'
                                                '<br><p><h4>Type:</h4> number (int or float)</p>\n'
                                                '<br><p><h4>Limits:</h4> 0.0:1.0 </p>',
    '_shelx_hkl_file'                         : '<pre><h2>_shelx_hkl_file</h2>'
                                                'HKL file used to refine the model in SHELXL. Do not '
                                                'edit this value!</pre>'
                                                '<br><p><h4>Type:</h4> string</p>\n',
    '_shelx_res_file'                         : '<pre><h2>_shelx_res_file</h2>'
                                                'Results file used to refine the model in SHELXL. Do not '
                                                'edit this value!</pre>'
                                                '<br><p><h4>Type:</h4> string</p>\n',
    '_shelx_hkl_checksum'                     : '<pre><h2>_shelx_hkl_checksum</h2>'
                                                'Checksum used to ensure the consistency of refinement model '
                                                'with the hkl data attached to the CIF.</pre>'
                                                '<br><p><h4>Type:</h4> integer number</p>\n',
    '_shelx_res_checksum'                     : '<pre><h2>_shelx_res_checksum</h2>'
                                                'Checksum used to ensure the consistency of refinement model '
                                                'with the results file attached to the CIF.</pre>'
                                                '<br><p><h4>Type:</h4> integer number</p>\n',
    '_olex2_diffrn_ambient_temperature_device': '<pre>Device to cool the '
                                                'crystal during measurement</pre>'
                                                '<h3>Example:</h3>\n'
                                                'Oxford Cryostream 800'
                                                '<br><p><h4>Type:</h4> string</p>',
    '_diffrn_measurement_ambient_'
    'temperature_device_make'                 : '<pre><h2>_diffrn_measurement_ambient_temperature_device_make</h2>'
                                                'Device model used to cool the crystal during the '
                                                'diffraction experiment,</pre>'
                                                '<h3>Example:</h3>\n'
                                                'Oxford Cryostream 800'
                                                '<br><p><h4>Type:</h4> string</p>',
    '_bruker_diffrn_'
    'measurement_temperature_device'          : '<pre><h2>_bruker_diffrn_measurement_temperature_device</h2>'
                                                'Device model used to cool the crystal during the '
                                                'diffraction experiment.</pre>'
                                                '<h3>Example:</h3>\n'
                                                'Oxford Cryostream 800'
                                                '<br><p><h4>Type:</h4> string</p>',
    '_diffrn_measurement_'
    'bruker_total_exposure_time'              : '<pre><h2>_diffrn_measurement_bruker_total_exposure_time</h2>'
                                                'Total X-ray exposure time of the crystal during data '
                                                'collection (seconds).</pre>'
                                                '<br><p><h4>Type:</h4> number (int or float)</p>',
    '_computing_bruker_data_scaling'          : '<pre><h2>_computing_bruker_data_scaling</h2>'
                                                'Program used to scale and correct the data for absorption.</pre>'
                                                '<br><p><h4>Type:</h4> Text</p>\n',
    '_diffrn_reflns_bruker_av_norm_I'         : '<pre><h2>_diffrn_reflns_bruker_av_norm_I</h2>'
                                                'Average normalized (to 1 min/deg) intensity.</pre>'
                                                '<br><p><h4>Type:</h4> number (float)</p>\n',
    '_diffrn_reflns_bruker_twinabs_number'    : '<pre><h2>_diffrn_reflns_bruker_twinabs_number</h2>'
                                                'Total number of measured intensities as counted by TWINABS.</pre>'
                                                '<br><p><h4>Type:</h4> number (float)</p>\n',
    '_platon_squeeze_void_nr'                 : '<pre><h2>_platon_squeeze_void_nr</h2>'
                                                'Sequential number for the voids where SQUEEZE found electron desnity.</pre>'
                                                '<br><p><h4>Type:</h4> number (int)</p>\n',
    '_platon_squeeze_void_average_x'          : '<pre><h2>_platon_squeeze_void_average_x</h2>'
                                                'x-position of the void.</pre>'
                                                '<br><p><h4>Type:</h4> number (float)</p>\n',
    '_platon_squeeze_void_average_y'          : '<pre><h2>_platon_squeeze_void_average_y</h2>'
                                                'y-position of the void.</pre>'
                                                '<br><p><h4>Type:</h4> number (float)</p>\n',
    '_platon_squeeze_void_average_z'          : '<pre><h2>_platon_squeeze_void_average_z</h2>'
                                                'z-position of the void.</pre>'
                                                '<br><p><h4>Type:</h4> number (float)</p>\n',
    '_platon_squeeze_void_volume'             : '<pre><h2>_platon_squeeze_void_volume</h2>'
                                                'Volume of the respective solvent mask void.</pre>'
                                                '<br><p><h4>Type:</h4> number (float)</p>\n',
    '_platon_squeeze_void_count_electrons'    : '<pre><h2>_platon_squeeze_void_count_electrons</h2>'
                                                'Number of electrons representing the molecular content of the void.</pre>'
                                                '<br><p><h4>Type:</h4> number (float)</p>\n',
    '_platon_squeeze_void_content'            : '<pre><h2>_platon_squeeze_void_content</h2>'
                                                'Formula that describes the content of the void as close as possible.'
                                                'The _chemical_formula_sum should not contain the SQUEEZEd atoms.</pre>'
                                                '<br><p><h4>Type:</h4> string </p>\n',
    '_platon_squeeze_void_probe_radius'       : '<pre><h2>_platon_squeeze_void_probe_radius</h2>'
                                                'Probe radius used to determine the void volume.</pre>'
                                                '<br><p><h4>Type:</h4> number (float) </p>\n',
    '_platon_squeeze_details'                 : '<pre><h2>_platon_squeeze_details</h2>'
                                                'Detailed description of the SQUEEZE process and the reasons why it '
                                                'was chosen over an atomistic (disorder) model.</pre>'
                                                '<h3>Example:</h3>\n'
                                                'The O(CH2CH3)2 solvent molecule (42 electrons) disordred around '
                                                'an inverstion center was treated by SQUEEZE. '
                                                '<br><p><h4>Type:</h4> string </p>\n',
    '_exptl_bruker_absorpt_correction_T_ratio': '<pre><h2>_exptl_bruker_absorpt_correction_T_ratio</h2>'
                                                'SADABS (and other scaling programs) can only determine the RATIO of '
                                                'absorpt_correction_Tmin/Tmax, so this value determined by SHELXL is used to '
                                                'scale the transmission values from SADABS.</pre>'
                                                '<br><p><h4>Type:</h4> number (float) </p>\n',
}

cif_all_dict = {}

cif_all_dict.update(cif_core)
cif_all_dict.update(twinning_dict)
cif_all_dict.update(modulation_dict)
cif_all_dict.update(powder_dict)
cif_all_dict.update(restraints_dict)
cif_all_dict.update(additional_keywords)



#  ----------------------------------------------------------------------------
#  "THE BEER-WARE LICENSE" (Revision 42):
#  dkratzert@gmx.de> wrote this file.  As long as you retain
#  this notice you can do whatever you want with this stuff. If we meet some day,
#  and you think this stuff is worth it, you can buy me a beer in return.
#  Dr. Daniel Kratzert
#  ----------------------------------------------------------------------------
import os
import re
from collections import namedtuple
from collections.abc import Generator
from contextlib import suppress
from pathlib import Path

import gemmi

from finalcif.cif import atoms

if hasattr(gemmi, 'set_leak_warnings'):
    gemmi.set_leak_warnings(False)

from gemmi.cif import as_string, Document, Loop
from packaging.version import Version
from shelxfile import Shelxfile

from finalcif.cif.cif_order import order, special_keys, essential_keys, non_centrosymm_keys
from finalcif.cif.hkl import HKL, Limit
from finalcif.cif.text import utf8_to_str, quote, retranslate_delimiter
from finalcif.datafiles.utils import DSRFind
from finalcif.tools.misc import isnumeric, grouper, strip_finalcif_of_name, _angstrom_to_x


class GemmiError(Exception):
    pass


class CifContainer:
    """
    This class holds the content of a cif file, independent of the file parser used.
    """

    def __init__(self, file: Path | str, new_block: str = ''):
        """

        Args:
            file: CIF file to open
            new_block: Create a new block (new file) if a name is given. Otherwise, just
                       the existing document is opened.
        """
        self.chars_ok = True
        self.picometer: bool = False
        self._abs_hkl_details: dict[str, str] = {}
        if isinstance(file, str):
            self.fileobj = Path(file)
        elif isinstance(file, Path):
            self.fileobj = file
        else:
            raise TypeError('The file parameter must be string or Path object.')
        self.current_block = ''
        # I do this in small steps instead of gemmi.cif.read_file() in order to
        # leave out the check_for_missing_values. This way, gemmi reads cif files even
        # with missing values.
        if new_block:
            self.doc: Document = gemmi.cif.Document()
            self.doc.add_new_block(new_block)
        else:
            try:
                self.doc = self.read_file(self.fileobj)
            except Exception as e:
                raise GemmiError(e)
        # Starting with first block, but can use others with subsequent self._onload():
        self.block: gemmi.cif.Block = self.doc[0]
        self.shx = Shelxfile(verbose=True)
        self.shx.read_string(self.res_file_data[1:-1])
        self._on_load()
        self.order = []
        self.essential_keys = []
        # May be overwritten later:
        self.set_order_keys(order)
        self.set_essential_keys(essential_keys)

    def set_order_keys(self, keys: list[str]) -> None:
        self.order = keys

    def set_essential_keys(self, keys: list[str]) -> None:
        self.essential_keys = keys

    @property
    def path_base(self) -> Path:
        """
        The absolute path of the current file without file name:
        /foo/bar/baz
        """
        return Path(self.doc.source).resolve().parent

    @property
    def filename(self) -> str:
        """
        The name of the current file without path:
        foo.cif
        """
        return Path(self.doc.source).name or self.fileobj.name

    @property
    def finalcif_file(self) -> Path:
        """
        The full path of the file with '-finalcif' attached to the end:
        foo/bar/baz-finalcif.cif
        """
        filename = self.finalcif_file_prefixed(prefix='', suffix='-finalcif.cif')
        return filename

    def finalcif_file_prefixed(self, prefix: str, suffix: str = '-finalcif.cif', force_strip=True) -> Path:
        """
        The full path of the file with a prefix and '-finalcif.cif' attached to the end.
        The suffix needs '-finalcif' in order to contain the finalcif ending.
        "foo/bar/baz-finalcif.cif"

        :param force_strip: Forces to strip the filename also after the '-finalcif' string.
        """
        file_witout_finalcif = strip_finalcif_of_name(Path(self.filename).stem, till_name_ends=force_strip)
        filename = self.path_base.joinpath(Path(prefix + file_witout_finalcif + suffix))
        return filename

    def get_line_numbers_of_bad_characters(self, filepath: Path):
        line_numbers = []
        for num, line in enumerate(filepath.read_bytes().splitlines(keepends=True)):
            try:
                line.decode('ascii')
            except UnicodeDecodeError:
                line_numbers.append(num)
        return line_numbers

    @property
    def is_multi_cif(self) -> bool:
        return True if len(self.doc) > 1 else False

    def load_this_block(self, index: int) -> None:
        self.block = self.doc[index]
        self.current_block = self.block.name
        self._on_load()

    def load_block_by_name(self, blockname: str) -> None:
        self.block = self.doc.find_block(blockname)
        self.current_block = self.block.name
        self._on_load()

    def delete_block(self, index: int) -> None:
        del self.doc[index]
        self.block: gemmi.cif.Block = self.doc[0]

    def _on_load(self) -> None:
        # will not ok with non-ascii characters in the res file:
        self.chars_ok = True
        try:
            self.doc.check_for_duplicates()
        except RuntimeError as e:
            raise GemmiError(e)
        self.dsr_used = DSRFind(self.res_file_data).dsr_used
        self.atomic_struct: gemmi.SmallStructure = gemmi.make_small_structure_from_block(self.block)
        # A dictionary to convert Atom names like 'C1_2' or 'Ga3' into Element names like 'C' or 'Ga'
        self._name2elements = dict(zip([x.upper() for x in self.block.find_loop('_atom_site_label')],
                                       [y.upper() for y in self.block.find_loop('_atom_site_type_symbol')], strict=False))
        self.check_hkl_min_max()

    @property
    def is_valid_structure_cif(self):
        """
        Checkl wether this CIF contains a structure or just metadata.
        """
        return len(self.atomic_struct.sites) > 0 and self.cell.a

    @property
    def hkl_extra_info(self):
        if not self._abs_hkl_details:
            self._abs_hkl_details = self._sadabs_hkl_details()
        return self._abs_hkl_details

    def check_hkl_min_max(self) -> None:
        if not all([self['_diffrn_reflns_limit_h_min'], self['_diffrn_reflns_limit_h_max'],
                    self['_diffrn_reflns_limit_k_min'], self['_diffrn_reflns_limit_k_max'],
                    self['_diffrn_reflns_limit_l_min'], self['_diffrn_reflns_limit_l_max']]) and self.hkl_file:
            limits = self.min_max_diffrn_reflns_limit()
            self['_diffrn_reflns_limit_h_min'] = str(limits.h_min)
            self['_diffrn_reflns_limit_h_max'] = str(limits.h_max)
            self['_diffrn_reflns_limit_k_min'] = str(limits.k_min)
            self['_diffrn_reflns_limit_k_max'] = str(limits.k_max)
            self['_diffrn_reflns_limit_l_min'] = str(limits.l_min)
            self['_diffrn_reflns_limit_l_max'] = str(limits.l_max)

    def read_file(self, path: Path) -> gemmi.cif.Document:
        """
        Reads a cif file and returns a gemmi document object.
        """
        doc = gemmi.cif.Document()
        # support for platon squeeze files:
        if path.suffix == '.sqf':
            txt = path.read_text(encoding='ascii')
            txt = 'data_justrandomlkdsadflkmcn\n' + txt
            doc.parse_string(txt)
        else:
            doc.source = str(path.resolve())
            doc.parse_file(str(path.resolve()))
        return doc

    def read_string(self, cif_string: str) -> gemmi.cif.Document:
        """
        Reads a cif file from a string and returns a gemmi cif docment.
        :param cif_string: cif as string
        :return: gemmi document
        """
        doc = gemmi.cif.Document()
        doc.parse_string(cif_string)
        return doc

    def cif_as_string(self, without_hkl=False) -> str:
        opt = gemmi.cif.WriteOptions()
        opt.align_pairs = 33
        if without_hkl:
            # return a copy, do not delete hkl from original:
            doc = gemmi.cif.Document()
            doc.parse_string(self.doc.as_string(options=opt))
            for block in doc:
                if block.find_pair_item('_shelx_hkl_file'):
                    block.find_pair_item('_shelx_hkl_file').erase()
                if block.find_pair_item('_shelx_fcf_file'):
                    block.find_pair_item('_shelx_fcf_file').erase()
            return doc.as_string(options=opt)
        else:
            return self.doc.as_string(options=opt)

    def __getitem__(self, item: str) -> str:
        """
        This method returns an empty string when the item value is '?'
        """
        if value := self.block.find_value(item):
            return as_string(value)
        else:
            return ''

    def __setitem__(self, key: str, value: str) -> None:
        """Set a key value pair of the current block.
        Values are automatically encoded from utf-8 and delimited.
        """
        self.set_pair_delimited(key, value)

    def __delitem__(self, key: str) -> None:
        with suppress(AttributeError):
            self.block.find_pair_item(key).erase()

    def __contains__(self, item) -> bool:
        return bool(self.__getitem__(item))

    def __str__(self) -> str:
        return (f"CIF file: {self.fileobj.resolve()!s}\n"
                f"{len(self.doc)} Block(s): {', '.join([x.name for x in self.doc])}\n"
                f"Contains SHELX res file: {True if self.res_file_data else False}\n"
                f"Has {self.natoms()} atoms"
                f", {self.nbonds()} bonds"
                f", {self.nangles()} angles")

    def file_is_there_and_writable(self) -> bool:
        return self.fileobj.exists() and self.fileobj.is_file() and os.access(self.fileobj, os.W_OK)

    def set_pair_delimited(self, key: str, txt: str):
        """
        Converts special characters to their markup counterparts.
        """
        txt = utf8_to_str(txt)
        try:
            # bad hack to get the numbered values correct
            float(txt)
            self.block.set_pair(key, txt)
        except (TypeError, ValueError):
            # prevent _key '?' in cif:
            if txt == '?':
                self.block.set_pair(key, txt)
            else:
                self.block.set_pair(key, quote(txt))

    def save(self, filename: Path | None = None) -> None:
        """
        Saves the current cif file. It uses the '[basename]-finalcif.cif' suffix by default.
        :param filename:  Name to save cif file to if the default is not sufficient.
        """
        if not filename:
            filename = self.finalcif_file
        if self.is_empty():
            print(f'DBG> CIF file {filename} is empty.')
            return
        if not self.is_writable(filename):
            raise PermissionError(f'Failed to open {filename.resolve()} for writing: Operation not permitted')
        self.order_cif_keys()
        print('Saving CIF to', Path(filename).resolve())
        if Version(gemmi.__version__) > Version('0.6.3'):
            options = gemmi.cif.WriteOptions()
            options.prefer_pairs = False
            options.compact = False
            options.align_pairs = 33
            options.align_loops = 15
            options.misuse_hash = False
            self.doc.write_file(str(filename), options=options)
        else:
            self.doc.write_file(str(filename), gemmi.cif.Style.Indent35)

    def is_writable(self, filepath: Path | str) -> bool:
        filepath = Path(filepath)
        try:
            exist = filepath.exists()
            with open(filepath, 'wb+') as f:
                f.write(b'')
            if not exist:
                Path(filepath).unlink(missing_ok=True)
            return True
        except Exception:
            return False

    def order_cif_keys(self) -> None:
        """
        Brings the current CIF in the specific order of the order list.
        """
        for block in self.doc:
            for key in reversed(self.order):
                try:
                    block.move_item(block.get_index(key), 0)
                except (RuntimeError, ValueError):
                    pass
                    # print('Not in list:', key)
            # make sure hkl file and res file are at the end if the cif file:
            for key in special_keys:
                try:
                    block.move_item(block.get_index(key), -1)
                except (RuntimeError, ValueError):
                    continue
            # Move all _vrf_* keys to the very top of the block:
            vrf_keys = [item.pair[0] for item in block if item.pair and item.pair[0].startswith('_vrf')]
            for key in reversed(vrf_keys):
                try:
                    block.move_item(block.get_index(key), 0)
                except (RuntimeError, ValueError):
                    pass

    @property
    def res_file_data(self) -> str:
        try:
            if self['_shelx_res_file']:
                return self['_shelx_res_file']
            elif self['_iucr_refine_instructions_details']:
                return self['_iucr_refine_instructions_details']
            else:
                return ''
        except UnicodeDecodeError:
            # This is a fallback in case _shelx_res_file has non-ascii characters.
            print('File has non-ascii characters. Switching to compatible mode.')
            self.doc = self.read_string(self.fileobj.read_text(encoding='latin1', errors='ignore'))
            self.block = self.doc.sole_block()
            self.chars_ok = False
            return self.block.find_value('_shelx_res_file')

    @property
    def hkl_file(self) -> str:
        hkl_loop: Loop = self.get_loop('_diffrn_refln_index_h')
        if hkl_loop and hkl_loop.width() > 4:
            return self._hkl_from_cif_format(hkl_loop)
        else:
            # returns an empty string if no cif hkl was found:
            return self._hkl_from_shelx()

    def _hkl_from_cif_format(self, hkl_loop: Loop) -> str:
        hkl_list = []
        format_string = '{:>4}{:>4}{:>4}{:>8}{:>8}'
        if hkl_loop.width() == 6:
            format_string = '{:>4}{:>4}{:>4}{:>8}{:>8}{:>4}'
        for line in grouper(hkl_loop.values, hkl_loop.width()):
            hkl_list.append(format_string.format(*line[:6]))
        return '\n'.join(hkl_list)

    def _hkl_from_shelx(self) -> str:
        try:
            if self['_shelx_hkl_file']:
                return self['_shelx_hkl_file']
            elif self['_iucr_refine_reflections_details']:
                return self['_iucr_refine_reflections_details']
            elif self['_xd_hkl_file']:
                return self['_xd_hkl_file']
            else:
                return ''
        except Exception as e:
            print(f'No hkl data found in CIF!, {e}')
            return ''

    @property
    def fab_file(self):
        if self['_shelx_fab_file']:
            return self['_shelx_fab_file']
        return ''

    @property
    def hkl_as_cif(self) -> str:
        return HKL(self.hkl_file, self.block.name, hklf_type=self.hklf_number).hkl_as_cif

    @property
    def hklf_number(self) -> int:
        hklf = 4
        if self.res_file_data:
            hklf = self._hklf_number_from_shelxl_file()
        return hklf

    def min_max_diffrn_reflns_limit(self) -> Limit:
        return HKL(self.hkl_file, self.block.name, hklf_type=self.hklf_number).get_hkl_min_max()

    def _hklf_number_from_shelxl_file(self) -> int:
        if self.shx and self.shx.hklf:
            return self.shx.hklf.n if self.shx.hklf.n != 0 else 4
        else:
            return 4

    @property
    def hkl_file_without_foot(self) -> str:
        """Returns a hkl file with no content after the 0 0 0 reflection"""
        zero_reflection_position = self._find_line_of_000(self.hkl_file)
        if zero_reflection_position:
            return '\n'.join(self.hkl_file.splitlines(keepends=False)[:zero_reflection_position + 1])
        else:
            return self.hkl_file

    @staticmethod
    def _find_line_of_000(lines: str) -> int:
        pattern = re.compile(r'^\s+0\s+0\s+0\s+0.*')
        for num, line in enumerate(lines.splitlines(keepends=False)):
            found = pattern.match(line)
            if found and num > 0:
                return num
        return 0

    def _sadabs_hkl_details(self) -> dict[str, str]:
        """
        This method tries to determine the information witten at the end of a cif hkl file by sadabs.
        """
        hkl = None
        all_sadabs_items = {'_exptl_absorpt_process_details' : '',
                            '_exptl_absorpt_correction_type' : '',
                            '_exptl_absorpt_correction_T_max': '',
                            '_exptl_absorpt_correction_T_min': '',
                            '_computing_structure_solution'  : '',
                            }
        try:
            hkl = self.hkl_file
        except Exception:
            pass
        if not hkl:
            return all_sadabs_items
        # zero_reflection_position = self.normal_search(hkl)
        hkl_splitted = hkl.splitlines(keepends=False)
        zero_reflection_position = self.reversed_search(hkl_splitted)
        if not zero_reflection_position:
            return all_sadabs_items
        hkl = hkl.splitlines(keepends=False)[zero_reflection_position:]
        # html-embedded cif has ')' instead of ';':
        hkl = [';' if x.startswith(')') else x for x in hkl]
        # the keys have a blank char in front:
        hkl = [re.sub(r'^ _', '_', x) for x in hkl]
        hkl = 'data_hkldat\n' + '\n'.join(hkl)
        try:
            hkldoc = gemmi.cif.read_string(hkl)
            hklblock = hkldoc.sole_block()
        except Exception as e:
            print('Unable to get information from hkl foot.')
            print(e)
            return all_sadabs_items
        for key in all_sadabs_items:
            val = hklblock.find_value(key)
            if val:
                all_sadabs_items[key] = gemmi.cif.as_string(val).strip()
        return all_sadabs_items

    def normal_search(self, hkl) -> int:
        pattern = re.compile(r'\s+0\s+0\s+0\s+0')
        found = pattern.search(hkl)
        if found:
            zero_reflection_position = found.start()
        else:
            zero_reflection_position = 0
        return zero_reflection_position

    def reversed_search(self, hkl_splitted: list[str]) -> int:
        pattern = re.compile(r'\s+0\s+0\s+0\s+0')
        zero_reflection_position = 0
        for num, line in enumerate(reversed(hkl_splitted)):
            found = pattern.search(line)
            if num > 500:
                # A longer footer is not realistic
                break
            if found:
                zero_reflection_position = len(hkl_splitted) - num
        return zero_reflection_position

    def is_empty(self) -> bool:
        """
        Checks if the currebt CIF contains data, but this works only for single-CIFs!
        """
        if len(self.doc) < 2 and len(self.keys()) + len(self.loops) == 0:
            return True
        return False

    def keys(self) -> list[str]:
        """
        Returns a plain list of keys that are really in this CIF.
        """
        return [x.pair[0] for x in self.block if x.pair is not None]

    def values(self):
        """
        Returns a plain list of keys that are really in this CIF.
        """
        return [x.pair[1] for x in self.block if x.pair is not None]

    @property
    def loops(self) -> list[gemmi.cif.Loop]:
        """
        Returns a list of loops contained in the current block.
        """
        loops = []
        for b in self.block:
            if b.loop:
                loops.append(b.loop)
        return loops

    def add_loop_to_cif(self, loop_tags: list[str], row_values: list[str] | tuple[str]) -> gemmi.cif.Loop:
        gemmi_loop = self.init_loop(loop_tags)
        for row in list(grouper(row_values, len(loop_tags))):
            gemmi_loop.add_row(row)
        return gemmi_loop

    def add_loop_from_columns(self, loop_tags: list[str],
                              column_values: list[list[str]] | tuple[tuple[str]]) -> gemmi.cif.Loop:
        try:
            gemmi_loop = self.init_loop(loop_tags)
            gemmi_loop.set_all_values(column_values)
        except RuntimeError as err:
            raise Exception('Gemmi raised an exception') from err
        return gemmi_loop

    @property
    def n_loops(self) -> int:
        return len(self.loops)

    def get_loop(self, key_in_loop: str) -> gemmi.cif.Loop:
        return self.block.find_loop(key_in_loop).get_loop()

    def get_loop_column(self, key_in_loop: str) -> list:
        return [retranslate_delimiter(as_string(x)) for x in self.block.find_loop(key_in_loop)]

    def init_loop(self, loop_keywords: list) -> gemmi.cif.Loop:
        return self.block.init_loop('', loop_keywords)

    @property
    def z_value(self) -> float:
        return self.atomic_struct.cell.volume / self.atomic_struct.cell.volume_per_image()

    @property
    def solution_program_details(self) -> str:
        return self.hkl_extra_info['_computing_structure_solution']

    @property
    def absorpt_process_details(self) -> str:
        return self.hkl_extra_info['_exptl_absorpt_process_details']

    @property
    def absorpt_correction_type(self) -> str:
        return self.hkl_extra_info['_exptl_absorpt_correction_type']

    @property
    def absorpt_correction_t_max(self) -> str:
        return self.hkl_extra_info['_exptl_absorpt_correction_T_max']

    @property
    def absorpt_correction_t_min(self) -> str:
        return self.hkl_extra_info['_exptl_absorpt_correction_T_min']

    def _spgr(self) -> gemmi.SpaceGroup:
        if self.symmops and self.symmops != ['']:
            symm_ops = self.symmops
        else:
            symm_ops = self.symmops_from_spgr
        return gemmi.find_spacegroup_by_ops(gemmi.GroupOps([gemmi.Op(o) for o in symm_ops]))

    @property
    def space_group(self) -> str:
        """
        Returns the space group from the symmetry operators.
        spgr.short_name() gives the short name.
        """
        try:
            return self._spgr().xhm()
        except (AttributeError, RuntimeError, ValueError):
            if self['_space_group_name_H-M_alt']:
                return gemmi.cif.as_string(self['_space_group_name_H-M_alt'])
            elif self['_symmetry_space_group_name_H-M']:
                return gemmi.cif.as_string(self['_symmetry_space_group_name_H-M'])
            else:
                return ''

    @property
    def symmops_from_spgr(self) -> list[str]:
        # _symmetry_space_group_name_Hall
        space_group = None
        if self['_space_group_name_H-M_alt']:
            space_group = self['_space_group_name_H-M_alt']
        if self['_symmetry_space_group_name_H-M']:
            space_group = self['_symmetry_space_group_name_H-M']
        if not space_group:
            return []
        ops = [op.triplet() for op in
               gemmi.find_spacegroup_by_name(gemmi.cif.as_string(space_group)).operations()]
        return ops

    @property
    def spgr_number_from_symmops(self) -> int:
        return self._spgr().number

    @property
    def spgr_number(self) -> int:
        if self['_space_group_IT_number'] and isnumeric(self['_space_group_IT_number']):
            return int(self['_space_group_IT_number'])
        elif self['_symmetry_Int_Tables_number'] and isnumeric(self['_symmetry_Int_Tables_number']):
            return int(self['_symmetry_Int_Tables_number'])
        else:
            return self.spgr_number_from_symmops

    @property
    def crystal_system(self) -> str:
        if not self._spgr():
            return ''
        return self._spgr().crystal_system_str()

    @property
    def hall_symbol(self) -> str:
        return self._spgr().hall

    @property
    def hkl_checksum_calcd(self) -> int:
        """
        Calculates the shelx checksum for the hkl file content of a cif file.
        """
        if self.hkl_file:
            return self.calc_checksum(self.hkl_file)
        else:
            return 0

    @property
    def res_checksum_calcd(self) -> int:
        """
        Calculates the shelx checksum for the res file content of a cif file.
        """
        if self.res_file_data:
            return self.calc_checksum(self.res_file_data)
        return 0

    @staticmethod
    def calc_checksum(input_str: str) -> int:
        """
        Calculates the shelx checksum of a cif file.
        The original algorithm was posted by Berthold Stöger on the Bruker Users Mailing list.
        """
        crc_sum = 0
        try:
            bytes_str = input_str.encode('cp1250', 'ignore')
        except Exception:
            bytes_str = input_str.encode('ascii', 'ignore')
        for char in bytes_str:
            # print(char)
            if char > 32:  # ascii 32 is space character
                crc_sum += char
        crc_sum %= 714025
        crc_sum = crc_sum * 1366 + 150889
        crc_sum %= 714025
        crc_sum %= 100000
        return crc_sum

    def rename_data_name(self, newname: str = ''):
        """
        Renames data_ tags to the newname. Also _vrf tags are renamed accordingly.
        http://journals.iucr.org/services/cif/checking/checkfaq.html
        """
        newname = ''.join([i for i in newname if i.isascii()])
        self.block.name = newname
        for item in self.block:
            if item.pair is not None:
                key, value = item.pair
                if key.startswith('_vrf'):
                    newkey = '_' + '_'.join(key.split('_')[1:3]) + '_' + newname
                    self.block.find_pair_item(key).erase()
                    self.block.set_pair(newkey, value)

    @property
    def symmops(self) -> list[str]:
        """
        Reads the symmops from the cif file.
        """
        xyz1 = self.block.find(["_symmetry_equiv_pos_as_xyz"])  # deprecated
        xyz2 = self.block.find(["_space_group_symop_operation_xyz"])  # New definition
        if xyz1:
            return [i.str(0) for i in xyz1]
        elif xyz2:
            return [i.str(0) for i in xyz2]
        else:
            return self.symmops_from_spgr

    @property
    def is_centrosymm(self) -> bool:
        """
        Whether a structure is centrosymmetric or not.
        """
        if not self.symmops or self.symmops == ['']:
            # Do not crash without symmops
            return False
        ops = gemmi.GroupOps([gemmi.Op(o) for o in self.symmops])
        return ops.is_centrosymmetric()

    def atoms(self, without_h: bool = False) -> Generator:
        """
        Atoms from the CIF where values are returned as string like in the CIF with esds.
        """
        labels = self.block.find_loop('_atom_site_label')
        types = self.block.find_loop('_atom_site_type_symbol') or [atoms.get_atomlabel(x) for x in labels]
        x = self.block.find_loop('_atom_site_fract_x')
        y = self.block.find_loop('_atom_site_fract_y')
        z = self.block.find_loop('_atom_site_fract_z')
        part = self.block.find_loop('_atom_site_disorder_group') or ('0',) * len(labels)
        occ = self.block.find_loop('_atom_site_occupancy') or ('1.000000',) * len(labels)
        u_eq = self.block.find_loop('_atom_site_U_iso_or_equiv') or ('.',) * len(labels)
        atom = namedtuple('atom', ('label', 'type', 'x', 'y', 'z', 'part', 'occ', 'u_eq'))
        for label, type, x, y, z, part, occ, u_eq in zip(labels, types, x, y, z,
                                                         part,
                                                         occ,
                                                         u_eq, strict=False):
            if without_h and self.ishydrogen(label):
                continue
            if self.picometer:
                u_eq = _angstrom_to_x(u_eq, factor=100 ** 2)
            #         0    1   2  3  4   5   6     7
            # yield label, type, x, y, z, part, occ, ueq
            yield atom(label=label, type=type, x=x, y=y, z=z, part=part, occ=occ, u_eq=u_eq)

    @property
    def atoms_fract(self) -> Generator:
        """Atoms with numbers as float values without esd."""
        for at in self.atomic_struct.sites:
            yield [at.label, at.type_symbol, at.fract.x, at.fract.y, at.fract.z, at.disorder_group, at.occ, at.u_iso]

    @property
    def atoms_orth(self) -> Generator:
        atom = namedtuple('atom', ('label', 'type', 'x', 'y', 'z', 'part', 'occ', 'u_eq'))
        cell = self.atomic_struct.cell
        for at in self.atomic_struct.sites:
            x, y, z = at.orth(cell)
            yield atom(label=at.label, type=at.type_symbol, x=x, y=y, z=z,
                       part=at.disorder_group, occ=at.occ, u_eq=at.u_iso)

    def displacement_parameters(self):
        """
        Yields the anisotropic displacement parameters.
        """
        labels = self.block.find_loop('_atom_site_aniso_label')
        u11 = self.block.find_loop('_atom_site_aniso_U_11')
        u22 = self.block.find_loop('_atom_site_aniso_U_22')
        u33 = self.block.find_loop('_atom_site_aniso_U_33')
        u23 = self.block.find_loop('_atom_site_aniso_U_23')
        u13 = self.block.find_loop('_atom_site_aniso_U_13')
        u12 = self.block.find_loop('_atom_site_aniso_U_12')
        adp = namedtuple('adp', ('label', 'U11', 'U22', 'U33', 'U23', 'U13', 'U12'))
        for label, u11, u22, u33, u23, u13, u12 in zip(labels, u11, u22, u33, u23, u13, u12, strict=True):
            if self.picometer:
                u11 = _angstrom_to_x(u11, factor=100 ** 2)
                u22 = _angstrom_to_x(u22, factor=100 ** 2)
                u33 = _angstrom_to_x(u33, factor=100 ** 2)
                u23 = _angstrom_to_x(u23, factor=100 ** 2)
                u13 = _angstrom_to_x(u13, factor=100 ** 2)
                u12 = _angstrom_to_x(u12, factor=100 ** 2)
            yield adp(label=label, U11=u11, U22=u22, U33=u33, U12=u12, U13=u13, U23=u23)

    @property
    def hydrogen_atoms_present(self) -> bool:
        for at in self.atomic_struct.sites:
            if at.type_symbol in ('H', 'D'):
                return True
        return False

    @property
    def disorder_present(self) -> bool:
        for at in self.atomic_struct.sites:
            if at.disorder_group:
                return True
        return False

    @property
    def cell(self):
        c = self.atomic_struct.cell
        cell = namedtuple('cell', 'a, b, c, alpha, beta, gamma, volume')
        return cell(c.a, c.b, c.c, c.alpha, c.beta, c.gamma, c.volume)

    def ishydrogen(self, label: str) -> bool:
        """
        Determines if an atom with the name 'label' is a hydrogen atom.
        """
        hydrogen = ('H', 'D')
        if self._iselement(label) in hydrogen:
            return True
        else:
            return False

    def checksymm(self, symm):
        """Add translation of 555 to symmetry elements without translation"""
        if len(symm) == 1 and symm not in {'.', '?'}:
            symm = symm + '_555'
        return symm

    def bonds(self, without_h: bool = False) -> Generator:
        """
        Yields a list of bonds in the cif file.
        """
        label1 = self.block.find_loop('_geom_bond_atom_site_label_1')
        label2 = self.block.find_loop('_geom_bond_atom_site_label_2')
        dist = self.block.find_loop('_geom_bond_distance')
        symm = self.block.find_loop('_geom_bond_site_symmetry_2')
        publ_loop = self.block.find_loop('_geom_bond_publ_flag') or len(label1) * ['?']
        bond = namedtuple('bond', ('label1', 'label2', 'dist', 'symm'))
        for label1, label2, dist, symm, publ in zip(label1, label2, dist, symm, publ_loop, strict=False):
            if (without_h and (self.ishydrogen(label1) or self.ishydrogen(label2))) or self.yes_not_set(publ):
                continue
            else:
                if self.picometer:
                    dist = _angstrom_to_x(dist)
                yield bond(label1=label1, label2=label2, dist=dist, symm=self.checksymm(symm))

    def bond_dist(self, pair: str) -> str | None:
        for p in self.bonds():
            if f'{p.label1}-{p.label2}' == pair:
                if self.picometer:
                    return _angstrom_to_x(p.dist)
                else:
                    return p.dist
        return None

    def angle(self, atoms: str) -> str | None:
        for p in self.angles():
            if f'{p.label1}-{p.label2}-{p.label3}' == atoms:
                return p.angle_val
        return None

    def torsion(self, atoms: str) -> str | None:
        for p in self.torsion_angles():
            if f'{p.label1}-{p.label2}-{p.label3}-{p.label4}' == atoms:
                return p.torsang
        return None

    def angles(self, without_H: bool = False) -> Generator:
        label1 = self.block.find_loop('_geom_angle_atom_site_label_1')
        label2 = self.block.find_loop('_geom_angle_atom_site_label_2')
        label3 = self.block.find_loop('_geom_angle_atom_site_label_3')
        angle_val = self.block.find_loop('_geom_angle')
        symm1 = self.block.find_loop('_geom_angle_site_symmetry_1')
        symm2 = self.block.find_loop('_geom_angle_site_symmetry_3')
        publ_loop = self.block.find_loop('_geom_angle_publ_flag') or len(label1) * ['?']
        angle = namedtuple('angle', ('label1', 'label2', 'label3', 'angle_val', 'symm1', 'symm2'))
        for label1, label2, label3, angle_val, symm1, symm2, publ in (
                zip(label1, label2, label3, angle_val, symm1, symm2, publ_loop, strict=False)):
            if ((without_H and (self.ishydrogen(label1) or self.ishydrogen(label2) or self.ishydrogen(label3))) or
                    self.yes_not_set(publ)):
                continue
            else:
                yield angle(label1=label1, label2=label2, label3=label3, angle_val=angle_val,
                            symm1=self.checksymm(symm1), symm2=self.checksymm(symm2))

    def _iselement(self, name: str) -> str:
        try:
            return self._name2elements[name.upper()]
        except KeyError:
            return atoms.get_atomlabel(name)

    def natoms(self, without_h: bool = False) -> int:
        return len(tuple(self.atoms(without_h)))

    def nbonds(self, without_h: bool = False) -> int:
        """
        Number of bonds in the cif object, with and without hydrogen atoms.
        """
        return len(tuple(self.bonds(without_h)))

    def nangles(self, without_h: bool = False) -> int:
        """
        Number of bond angles in the cif object, with and without hydrogen atoms.
        """
        return len(tuple(self.angles(without_h)))

    def ntorsion_angles(self, without_h: bool = False) -> int:
        """
        Number of torsion angles in the cif object, with and without hydrogen atoms.
        """
        return len(tuple(self.torsion_angles(without_h)))

    def torsion_angles(self, without_h: bool = False) -> Generator:
        label1 = self.block.find_loop('_geom_torsion_atom_site_label_1')
        label2 = self.block.find_loop('_geom_torsion_atom_site_label_2')
        label3 = self.block.find_loop('_geom_torsion_atom_site_label_3')
        label4 = self.block.find_loop('_geom_torsion_atom_site_label_4')
        torsang = self.block.find_loop('_geom_torsion')
        symm1 = self.block.find_loop('_geom_torsion_site_symmetry_1')
        symm2 = self.block.find_loop('_geom_torsion_site_symmetry_2')
        symm3 = self.block.find_loop('_geom_torsion_site_symmetry_3')
        symm4 = self.block.find_loop('_geom_torsion_site_symmetry_4')
        publ_loop = self.block.find_loop('_geom_torsion_publ_flag') or len(label1) * ['?']
        tors = namedtuple('Torsion',
                          ('label1', 'label2', 'label3', 'label4', 'torsang', 'symm1', 'symm2', 'symm3', 'symm4'))
        for label1, label2, label3, label4, torsang, symm1, symm2, symm3, symm4, publ in zip(label1, label2, label3,
                                                                                             label4,
                                                                                             torsang, symm1, symm2,
                                                                                             symm3,
                                                                                             symm4, publ_loop,
                                                                                             strict=False):
            if ((without_h and (self.ishydrogen(label1) or self.ishydrogen(label2)
                                or self.ishydrogen(label3) or self.ishydrogen(label3))) or self.yes_not_set(publ)):
                continue
            yield tors(label1=label1, label2=label2, label3=label3, label4=label4, torsang=torsang,
                       symm1=self.checksymm(symm1),
                       symm2=self.checksymm(symm2),
                       symm3=self.checksymm(symm3),
                       symm4=self.checksymm(symm4))

    def hydrogen_bonds(self) -> Generator:
        label_d = self.block.find_loop('_geom_hbond_atom_site_label_D')
        label_h = self.block.find_loop('_geom_hbond_atom_site_label_H')
        label_a = self.block.find_loop('_geom_hbond_atom_site_label_A')
        dist_dh = self.block.find_loop('_geom_hbond_distance_DH')
        dist_ha = self.block.find_loop('_geom_hbond_distance_HA')
        dist_da = self.block.find_loop('_geom_hbond_distance_DA')
        angle_dha = self.block.find_loop('_geom_hbond_angle_DHA')
        symm = self.block.find_loop('_geom_hbond_site_symmetry_A')
        publ_loop = self.block.find_loop('_geom_hbond_publ_flag') or len(label_d) * ['?']
        hydr = namedtuple('HydrogenBond', ('label_d', 'label_h', 'label_a', 'dist_dh', 'dist_ha', 'dist_da',
                                           'angle_dha', 'symm'))
        for label_d, label_h, label_a, dist_dh, dist_ha, dist_da, angle_dha, symm, publ in (
                zip(label_d, label_h, label_a, dist_dh, dist_ha, dist_da, angle_dha, symm, publ_loop, strict=False)):
            if self.yes_not_set(publ):
                continue
            if self.picometer:
                dist_dh = _angstrom_to_x(dist_dh)
                dist_ha = _angstrom_to_x(dist_ha)
                dist_da = _angstrom_to_x(dist_da)
            yield hydr(label_d, label_h, label_a, dist_dh, dist_ha, dist_da, angle_dha, self.checksymm(symm))

    def yes_not_set(self, publ: str):
        return publ not in {'y', 'yes', '?', '.'} or publ in {'n', 'no'}

    def key_value_pairs(self) -> list[tuple[str, str]]:
        """
        Returns the key/value pairs of a cif file sorted by priority.
        """
        keys_without_values, keys_with_values = self.keys_with_essentials()
        return [*keys_without_values, ('These below are already in:', '---------------------'), *keys_with_values]

    def pairs(self):
        for item in self.block:
            if item.pair is not None:
                yield item.pair

    def _is_centrokey(self, key) -> bool:
        """
        Is True if the kurrent key is only valid
        for non-centrosymmetric structures
        """
        return self.is_centrosymm and key in non_centrosymm_keys

    def keys_with_essentials(self) -> tuple[list[tuple[str, str]], list[tuple[str, str]]]:
        """
        Returns the keys to be displayed in the main table as two separate lists.
        """
        questions = []
        # contains the answered keys:
        with_values = []
        for item in self.block:
            if item.pair is not None:
                key, value = item.pair
                # if key.startswith('_shelx'):
                #    # No-one should edit shelx values:
                #    continue
                # if key == '_iucr_refine_instructions_details':
                #    continue
                if self._is_centrokey(key):
                    continue
                if not value or value in {'?', "'?'"}:
                    questions.append((key, value))
                else:
                    with_values.append((key, value))
        all_keys = [x[0] for x in with_values] + [x[0] for x in questions]
        self.check_for_missing_essential_keys(all_keys, questions)
        vrf_questions = [x for x in questions if x[0].startswith('_vrf')]
        non_vrf_questions = [x for x in questions if not x[0].startswith('_vrf')]
        vrf_with_values = [x for x in with_values if x[0].startswith('_vrf')]
        non_vrf_with_values = [x for x in with_values if not x[0].startswith('_vrf')]
        return [*vrf_questions, *vrf_with_values, *sorted(non_vrf_questions)], [*sorted(non_vrf_with_values)]

    def check_for_missing_essential_keys(self, all_keys: list[tuple[str, str]],
                                         questions: list[tuple[str, str]]) -> None:
        """
        Check if there are keys not in the cif but in essential_keys and append them if so.
        """
        for key in self.essential_keys:
            if key not in all_keys:
                if self._is_centrokey(key):
                    continue
                questions.append((key, '?'))
                self.block.set_pair(key, '?')

    def test_res_checksum(self) -> bool:
        """
        A method to check whether the checksums in the cif file fit to the content.
        """
        cif_res_ckecksum = 0
        res_checksum_calcd = self.res_checksum_calcd
        if res_checksum_calcd > 0:
            cif_res_ckecksum = self['_shelx_res_checksum'] or -1
            try:
                cif_res_ckecksum = int(cif_res_ckecksum)
            except ValueError:
                cif_res_ckecksum = -1
        if cif_res_ckecksum > 0 and cif_res_ckecksum != res_checksum_calcd:
            return False
        else:
            return True

    def test_hkl_checksum(self) -> bool:
        """
        A method to check whether the checksums in the cif file fit to the content.
        """
        cif_hkl_ckecksum = 0
        hkl_checksum_calcd = self.hkl_checksum_calcd
        if hkl_checksum_calcd > 0:
            cif_hkl_ckecksum = self['_shelx_hkl_checksum'] or -1
            try:
                cif_hkl_ckecksum = int(cif_hkl_ckecksum)
            except ValueError:
                cif_hkl_ckecksum = -1
        if cif_hkl_ckecksum > 0 and cif_hkl_ckecksum != hkl_checksum_calcd:
            return False
        else:
            return True


if __name__ == '__main__':
    # c = CifContainer('../41467_2015.cif')
    # c = CifContainer('test-data/p21c.cif')

    c = CifContainer('test-data/DK_Zucker2_0m.cif')
    c.load_this_block(len(c.doc) - 1)
    # pp(list(c.torsion_angles()))
    print(c.bond_dist('C1-C2'))
    # pp(c.hkl_extra_info)
    # print(c.hkl_file)
    # print(c.hkl_as_cif)
    # print(c.test_hkl_checksum())
    # s = Shelxfile()
    # print(c.res_file_data)
    # s.read_string(c.res_file_data)
    # print(s.hklf.n)

    # print(CifContainer('tests/examples/1979688.cif').hkl_as_cif[-250:])

    """
    import n#umpy as np
    mtz = gemmi.Mtz(with_base=True)
    data = [x.split() for x in c.hkl_file.splitlines()]
    mtz.add_column('I', 'I')
    mtz.add_column('SIGI', 'Q')
    mtz.set_data(data)
    mtz.set_cell_for_all(gemmi.UnitCell(c.cell.a, c.cell.b, c.cell.c, c.cell.alpha, c.cell.beta, c.cell.gamma))
    mtz.spacegroup = gemmi.find_spacegroup_by_name(c.space_group)
    mtz.update_reso()
    size = mtz.get_size_for_hkl()
    np.seterr(divide='ignore', invalid='ignore')
    ios =  intensity.array / sigma.array
    
    doc = gemmi.cif.read('tests/examples/1979688-finalcif.fcf')
    rblock = gemmi.hkl_cif_as_refln_block(doc[0])
    """



from __future__ import annotations

import re

from qtpy import QtCore, QtGui
from qtpy.QtCore import QEvent, QObject, Qt
from qtpy.QtGui import QColor, QKeySequence, QBrush
from qtpy.QtWidgets import (
    QAbstractScrollArea, QTableView, QWidget, QApplication,
    QHeaderView, QStyledItemDelegate, QStyleOptionViewItem,
)

from finalcif.cif.text import retranslate_delimiter
from finalcif.gui.cif_table_model import CifTableModel, Column, blue, CifRowData
from finalcif.gui.combobox import MyComboBox
from finalcif.gui.dialogs import show_keyword_help
from finalcif.gui.plaintextedit import MyQPlainTextEdit

DEBUG = False


class SeparatorWidget(QWidget):
    """A simple widget that paints the blue diagonal cross pattern for the separator row."""

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        brush = QBrush(blue)
        brush.setStyle(Qt.BrushStyle.DiagCrossPattern)
        painter.fillRect(self.rect(), brush)
        painter.end()


class CifTableDelegate(QStyledItemDelegate):
    """Delegate for painting separator rows and providing size hints."""

    def paint(self, painter, option: QStyleOptionViewItem, index):
        model = index.model()
        if model:
            row_data = model.get_row_data(index.row())
            if row_data and row_data.is_separator:
                painter.save()
                brush = QBrush(blue)
                brush.setStyle(Qt.BrushStyle.DiagCrossPattern)
                painter.fillRect(option.rect, brush)
                painter.restore()
                return
        super().paint(painter, option, index)

    def sizeHint(self, option: QStyleOptionViewItem, index) -> QtCore.QSize:
        model = index.model()
        if model:
            row_data = model.get_row_data(index.row())
            if row_data and row_data.is_separator:
                return QtCore.QSize(100, 10)
        return super().sizeHint(option, index)


class CifTableView(QTableView):
    """QTableView-based main CIF table, replacing MyCifTable(QTableWidget).

    Uses a :class:`CifTableModel` as its data store and persistent
    :class:`MyQPlainTextEdit` / :class:`MyComboBox` widgets placed via
    ``setIndexWidget()`` for each cell.
    """

    row_deleted = QtCore.Signal(str)
    textTemplate = QtCore.Signal(int)
    new_key = QtCore.Signal(str)

    def __init__(self, parent: QWidget = None, *args, **kwargs):
        super().__init__(parent=parent)
        self._parent = parent
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)

        # Model & delegate
        self._model = CifTableModel(self)
        self.setModel(self._model)
        self.setItemDelegate(CifTableDelegate(self))

        # Visual config
        self.installEventFilter(self)
        self.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setSelectionBehavior(QTableView.SelectionBehavior.SelectRows)
        self.setShowGrid(True)
        self.setWordWrap(True)
        self.setSortingEnabled(False)

        # Shortcuts
        del_shortcut = QtGui.QShortcut(QKeySequence('Ctrl+Del'), self)
        del_shortcut.activated.connect(self._delete_current_row)

        # Vertical header
        vheader = self.verticalHeader()
        vheader.setSectionsClickable(True)
        vheader.sectionClicked.connect(self._vheader_section_click)

    # ------------------------------------------------------------------
    # Row / column counts
    # ------------------------------------------------------------------

    @property
    def rows_count(self) -> int:
        return self._model.rowCount()

    @property
    def columns_count(self) -> int:
        return self._model.columnCount()

    @property
    def vheaderitems(self) -> list[str]:
        """List of CIF keywords in row order."""
        return self._model.vheaderitems

    def rowCount(self) -> int:  # noqa: N802 – keep old name
        return self._model.rowCount()

    # ------------------------------------------------------------------
    # Row management
    # ------------------------------------------------------------------

    def add_row(self, key: str, position: int | None = None) -> int:
        """Insert a new row into the model. Returns the row index."""
        return self._model.add_row(key=key, position=position)

    def delete_row(self, row: int | None = None) -> None:
        if row is None:
            row = self.currentIndex().row()
        if row < 0 or row >= self._model.rowCount():
            return
        key = self._model.remove_row(row)
        if key:
            self.row_deleted.emit(key)

    def _delete_current_row(self) -> None:
        self.delete_row()

    def delete_content(self) -> None:
        """Remove all rows."""
        self._model.clear_all()

    def bulk_load(self, rows) -> None:
        """Load all rows at once — much faster than per-row add_row."""
        self._model.bulk_load(rows)

    def has_key(self, key: str) -> bool:
        return self._model.has_key(key)

    def row_from_key(self, key: str) -> int:
        return self._model.row_from_key(key)

    def add_separation_line(self, row_num: int) -> None:
        """Mark *row_num* as the blue diagonal-cross separator."""
        self._model.set_separator(row_num)
        for col in range(3):
            idx = self._model.index(row_num, col)
            self.setIndexWidget(idx, SeparatorWidget(self))
        self.resizeRowToContents(row_num)

    # ------------------------------------------------------------------
    # Cell widgets
    # ------------------------------------------------------------------

    def cellWidget(self, row: int, column: int) -> QWidget | None:
        return self.indexWidget(self._model.index(row, column))

    def setCellWidget(self, row: int, column: int, widget) -> None:
        keys = self._model.vheaderitems
        if row < len(keys):
            widget.cif_key = keys[row]
        if column in (Column.CIF, Column.DATA):
            widget.setUneditable()
        self.setIndexWidget(self._model.index(row, column), widget)

    def widget_from_key(self, key: str, column: int) -> QWidget | None:
        return self.cellWidget(self._model.row_from_key(key), column)

    # ------------------------------------------------------------------
    # Text access
    # ------------------------------------------------------------------

    def setText(self, key: str, column: int, txt: str,
                row: int | None = None, color: QColor | None = None) -> None:
        """Set text in a cell, creating a :class:`MyQPlainTextEdit` if needed."""
        if txt:
            txt = retranslate_delimiter(txt)

        if row is None:
            row = self._model.row_from_key(key)

        idx = self._model.index(row, column)
        existing = self.indexWidget(idx)

        if isinstance(existing, MyComboBox):
            existing.setText(txt)
            return

        if isinstance(existing, MyQPlainTextEdit):
            existing.setText(txt, color=color, column=column)
            existing.cif_key = key
        else:
            textedit = MyQPlainTextEdit(self)
            textedit.color = color
            textedit.cif_key = key
            textedit.templateRequested.connect(self.goto_template_page)
            textedit.new_key.connect(lambda x: self.new_key.emit(x))
            self.setCellWidget(row, column, textedit)
            textedit.setText(txt, color=color, column=column)
            if column in (Column.CIF, Column.DATA):
                textedit.setUneditable()

        widget = self.indexWidget(idx)

        if color:
            if widget and hasattr(widget, 'setBackground'):
                widget.setBackground(color)

        # Keep model in sync — store the widget's actual display text so that
        # the delegate's sizeHint (which reads from the model) matches the
        # truncated content shown for large values like HKL files.
        if widget and hasattr(widget, 'toPlainText'):
            self._model.set_cell_text(row, column, widget.toPlainText())
        else:
            self._model.set_cell_text(row, column, txt)
        if color:
            self._model.set_cell_color(row, column, color)
        self.resizeRowToContents(row)

    def text(self, row: int, column: int) -> str:
        """Read text from a cell, preferring the live widget content."""
        widget = self.indexWidget(self._model.index(row, column))
        if widget:
            if hasattr(widget, 'toPlainText'):
                txt = widget.toPlainText()
                if txt:
                    return txt
            if hasattr(widget, 'currentText'):
                txt = widget.currentText()
                if txt:
                    return txt
        return self._model.get_cell_text(row, column)

    def getText(self, row: int, col: int) -> str:  # noqa: N802
        return self.text(row, col)

    def getTextFromKey(self, key: str, col: int) -> str:  # noqa: N802
        return self.text(self._model.row_from_key(key), col)

    def vheader_text(self, row: int) -> str:
        h = self._model.headerData(row, Qt.Orientation.Vertical)
        return str(h) if h else ''

    # ------------------------------------------------------------------
    # Combobox helper
    # ------------------------------------------------------------------

    def add_property_combobox(self, data: list, row_num: int, key: str) -> None:
        combobox = MyComboBox(self)
        combobox.cif_key = key
        combobox.textTemplate.connect(self.goto_template_page)
        self.setCellWidget(row_num, Column.EDIT, combobox)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        for num, value in data:
            try:
                combobox.addItem(retranslate_delimiter(value), num)
            except TypeError:
                print('Bad value in property:', value)
                if DEBUG:
                    raise
                continue
        combobox.setCurrentIndex(0)

    # ------------------------------------------------------------------
    # Navigation / selection
    # ------------------------------------------------------------------

    def setCurrentCell(self, row: int, column: int) -> None:  # noqa: N802
        self.setCurrentIndex(self._model.index(row, column))

    def currentRow(self) -> int:  # noqa: N802
        return self.currentIndex().row()

    def goto_template_page(self, row: int) -> None:
        self.setCurrentCell(row, Column.EDIT)
        self.textTemplate.emit(self.currentRow())

    # ------------------------------------------------------------------
    # Search / filter
    # ------------------------------------------------------------------

    def search(self, searchtext: str) -> None:
        if not searchtext:
            for row in range(self._model.rowCount()):
                self.setRowHidden(row, False)
            return
        searchtext = searchtext.replace('.', '_')
        pattern = re.compile(f'.*{searchtext}.*', re.IGNORECASE)
        keys = self._model.vheaderitems
        matched = {k for k in keys if pattern.match(k)}
        for row in range(self._model.rowCount()):
            self.setRowHidden(row, keys[row] not in matched)

    # ------------------------------------------------------------------
    # Clipboard
    # ------------------------------------------------------------------

    def copy_vhead_item(self) -> None:
        row = self.currentIndex().row()
        keys = self._model.vheaderitems
        if 0 <= row < len(keys):
            QApplication.clipboard().setText(keys[row])

    # ------------------------------------------------------------------
    # Vertical header help
    # ------------------------------------------------------------------

    def _vheader_section_click(self, section: int) -> None:
        from finalcif.cif.all_cif_dicts import cif_all_dict
        keys = self._model.vheaderitems
        if section >= len(keys):
            return
        itemtext = keys[section]
        keyword_help = cif_all_dict.get(itemtext)
        if keyword_help:
            keyword_help = retranslate_delimiter(keyword_help, no_html_unescape=True)
            show_keyword_help(self._parent, keyword_help, itemtext)
        elif itemtext.startswith('_vrf_'):
            helptxt = ('<pre><h2>Validation Response Form (VRF)</h2>\n'
                       'The Validation Response Form is supplied in the checkCIF report for '
                       'problems that have triggered Alerts. It '
                       'provides a field for the author to respond to the alert. '
                       'The response is clearly visible to the review process. \n'
                       'Usually, only level A or B alerts need a VRF.')
            show_keyword_help(self._parent, helptxt, itemtext)
        else:
            show_keyword_help(self._parent, 'No help available for this key.', '')

    # ------------------------------------------------------------------
    # Column layout
    # ------------------------------------------------------------------

    def distribute_cif_main_table_columns_evenly(self) -> None:
        hh = self.horizontalHeader()
        for col in (Column.CIF, Column.DATA, Column.EDIT):
            hh.setSectionResizeMode(col, QHeaderView.ResizeMode.Stretch)
        hh.setAlternatingRowColors(True)
        self.verticalHeader().setAlternatingRowColors(True)

    # ------------------------------------------------------------------
    # Event handling
    # ------------------------------------------------------------------

    def eventFilter(self, widget: QObject, event: QEvent):
        if event.type() == QEvent.Type.KeyRelease:
            if event.key() == QtCore.Qt.Key.Key_Backtab:
                row = self.currentIndex().row()
                if row > 0:
                    self.setCurrentCell(row - 1, 2)
                return True
            if event.key() == QtCore.Qt.Key.Key_Tab:
                self.setCurrentCell(self.currentIndex().row(), 2)
                return True
        return QObject.eventFilter(self, widget, event)

    def resizeEvent(self, e: QtGui.QResizeEvent) -> None:
        QtCore.QTimer(self).singleShot(0, self.resizeRowsToContents)
        super().resizeEvent(e)


if __name__ == '__main__':
    app = QApplication([])
    tabv = CifTableView()
    tabv.bulk_load(
        [CifRowData(key=f'key{i}', cif_value='cifval', data_value='dataval', edit_value='editval') for i in range(20)])
    tabv.show()
    app.exec()



from __future__ import annotations
from textwrap import wrap
from typing import TYPE_CHECKING

from qtpy import QtCore, QtGui
from qtpy.QtCore import Qt, QObject, QEvent
from qtpy.QtWidgets import QComboBox, QSizePolicy

from finalcif.gui.validators import validators

light_red = QtGui.QColor(254, 191, 189)
white = QtGui.QColor(255, 255, 255)

if TYPE_CHECKING:
    from finalcif.gui.cif_table_view import CifTableView


class MyComboBox(QComboBox):
    """
    A special QComboBox with convenient methods to set the background color and other things.
    """
    textTemplate = QtCore.Signal(int)

    def __init__(self, parent: CifTableView = None):
        super().__init__(parent)
        self.parent: CifTableView = parent
        self.cif_key = ''
        self.color = white
        self.default_palette = self.palette()
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Expanding)
        self.setEditable(True)  # only editable as new template
        self.installEventFilter(self)
        self.actionDelete = QtGui.QAction("Delete Row", self)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)
        self.addAction(self.actionDelete)
        self.actionDelete.triggered.connect(self._delete_row)
        action_template = QtGui.QAction("Text Template", self)
        self.addAction(action_template)
        action_template.triggered.connect(lambda: self.textTemplate.emit(self.row))

    def __str__(self):
        return self.currentText()

    @property
    def row(self) -> int:
        return self.parent.row_from_key(self.cif_key)

    def _delete_row(self) -> None:
        self.parent.delete_row(self.row)

    def _on_create_template(self) -> None:
        self.textTemplate.emit(self.row)

    def eventFilter(self, widget: QObject, event: QEvent):
        """
        Event filter to ignore wheel events in comboboxes to prevent accidental changes to them.
        """
        if event.type() == QEvent.Wheel:  # and widget and not widget.hasFocus():
            event.ignore()
            return True
        return QObject.eventFilter(self, widget, event)

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        super().keyPressEvent(event)
        self.validate_text(self.currentText())

    def validate_text(self, text: str):
        validator = validators.get(self.cif_key, None)
        if validator and not validator.valid(text):
            self.setBadStyle()
            self.setToolTip(validator.help_text)
        else:
            self.setToolTip('')
            self.setRegularStyle()

    def setBadStyle(self) -> None:
        palette = self.palette()
        palette.setColor(self.backgroundRole(), light_red)
        self.setPalette(palette)

    def setRegularStyle(self) -> None:
        self.setPalette(self.default_palette)

    def getBackgroundColor(self) -> QtGui.QColor:
        palette = self.palette()
        background_color = palette.color(QtGui.QPalette.ColorRole.Base)
        return background_color

    def setUneditable(self):
        # noinspection PyUnresolvedReferences
        self.setFlags(self.flags() ^ QtCore.Qt.ItemFlag.ItemIsEditable)

    def setText(self, txt: str):
        self.setEditText('\n'.join(wrap(txt, width=80)))

    def addItem(self, *__args):
        text = '\n'.join(wrap(__args[0], width=80))
        super().addItem(text, __args[1])
